/* 
 * status.c --
 *
 *	Turns a status value into a string.
 *
 * @COPYRIGHT@
 */

#ifndef lint
static char rcsid[] = "$Header: /cvs/cvs/ftn/mirage/common/status.c,v 1.1 2001/12/12 23:03:31 bakers Exp $";
#endif /* not lint */

#include <status.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>
#include <errno.h>

typedef struct {
    char	*string;	// String associated with msg
    int		errnoValue;	// Corresponding errno
} MsgInfo;

typedef struct {
    char	*name;		// Module name
    MsgInfo	*info;		// Error information
    int		size;		// Number of messages
} StatusMsgs;


static MsgInfo swarmStatusMsgs[] = {
    {"unrecognized type of storage device", 	EINVAL},
    {"duplicate sid",				EINVAL},
    {"could not resolve address from name",	ENOENT},
    {"no such device",				ENODEV},
    {"service name too long", 			ENAMETOOLONG},
    {"incompatible interface",			EINVAL},
    {"not implemented",				ENOSYS},
    {"unavailable",				ENOENT},
    {"invalid FID",				EINVAL},
    {"invalid argument",			EINVAL},
    {"I/O error",			       	EIO},
    {"invalid service level",		       	EINVAL},
};

static MsgInfo devStatusMsgs[] = {
    {"no such device",				ENODEV},
    {"address beyond bounds",			EFAULT},
};

static MsgInfo statusStatusMsgs[] = {
    {"invalid status code",			EINVAL},
};

static MsgInfo logStatusMsgs[] = {
    {"fragment attributes are corrupted",	EIO},
    {"invalid version in fragment attributes",  EIO},
    {"bad magic number in log record",		EIO},
    {"bad magic number in log record section",	EIO},
    {"section not found",			ENOENT},
    {"set is full",				ENOSPC},
    {"duplicate set",				EEXIST},
    {"set is frozen",				EINVAL},
    {"duplicate record",			EEXIST},
    {"record not found",			ENOENT},
    {"record already committed",		EINVAL},
    {"service not found",			EINVAL},
    {"service already exists",			EEXIST},
};

static MsgInfo striperStatusMsgs[] = {
    {"no such sid", 				ENOENT},
    {"bad magic number in striper attributes",	EIO},
    {"I/O error",				EIO},
    {"out of stripes",				ENOMEM},
    {"no such fragment",			ENOENT},
};
static MsgInfo storageStatusMsgs[] = {
    {"marked fragment not found",		ENOENT},
    {"FID exists",				EEXIST},
    {"no space",				ENOSPC},
    {"no such FID",				ENOENT},
};

#define NUM_CODES(array) (sizeof(array) / sizeof(MsgInfo))
static StatusMsgs statusMsgs[] = {
    {"Errno",	NULL, 0},
    {"Swarm",	swarmStatusMsgs, NUM_CODES(swarmStatusMsgs)},
    {"Dev",	devStatusMsgs, NUM_CODES(devStatusMsgs)},
    {"Status",	statusStatusMsgs, NUM_CODES(statusStatusMsgs)},
    {"Log",	logStatusMsgs, NUM_CODES(logStatusMsgs)},
    {"Striper",	striperStatusMsgs, NUM_CODES(striperStatusMsgs)},
    {"Storage", storageStatusMsgs, NUM_CODES(storageStatusMsgs)},
};

static Status	ValidStatus(Status status, char **errorMsg);

/*
 *----------------------------------------------------------------------
 *
 * Status_Msg --
 *
 *	Returns the string corresponding to the given status.
 *
 * Results:
 *	Statically allocated string.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

char *
Status_Msg(status)
    Status	status;
{
    unsigned int	module = STATUS_MODULE(status);
    unsigned int	msg = STATUS_MSGNUM(status);
    Status		code;
    char		*error;

    code = ValidStatus(status, &error);
    if (code != SUCCESS) {
	return error;
    }
    /* 
     * Module 0 corresponds to errno values. Linux uses -errno a lot, which
     * corresponds to module 0xffff.
     */
    if (module == 0) {
	return strerror(status);
    } if (module == 0xffff) {
	return strerror(-status);
    } else if (status == FAILURE) {
	return "generic failure";
    } else {
	return statusMsgs[module].info[msg].string;
    }
}

/*
 *----------------------------------------------------------------------
 *
 * Status_ToErrno --
 *
 *	Converts a Swarm status into the most appropriate errno.
 *
 * Results:
 *	STATUS_BADSTATUS if status is invalid, SUCCESS otherwise
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

Status
Status_ToErrno(status, errnoValue)
    Status		status;		// Status to be converted.
    int			*errnoValue;	// Returned errno value.
{

    Status		code;
    unsigned int	module = STATUS_MODULE(status);
    unsigned int	msg = STATUS_MSGNUM(status);
    char		*error;

    code = ValidStatus(status, &error);
    if (code != SUCCESS) {
	return STATUS_BADSTATUS;
    }
    /* 
     * Module 0 corresponds to errno values. Linux uses -errno a lot,
     * which corresponds to module 0xffff
     */
    if (module == 0) {
	*errnoValue = status;
    } else if (module == 0xffff) {
	*errnoValue = -status;
    } else if (status == FAILURE) {
	*errnoValue = EINVAL;
    } else {
	*errnoValue = statusMsgs[module].info[msg].errnoValue;
    }
    return SUCCESS;
}

/*
 *----------------------------------------------------------------------
 *
 * ValidStatus --
 *
 *	Checks if the given status is valid.
 *
 * Results:
 *	SUCCESS or FAILURE
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

static Status
ValidStatus(status, errorMsg)
    Status	status;		// Status to be checked for validity
    char	**errorMsg;	// Error message
{
    unsigned int	module = STATUS_MODULE(status);
    unsigned int	msg = STATUS_MSGNUM(status);
    static char		buffer[100];

    /*
     * Make sure a Status is 32 bits.
     */
    assert(sizeof(Status) == 4);
    if (status == FAILURE) {
	return SUCCESS;
    }
    /* 
     * Module 0 corresponds to errno values.
     */
    if (module == 0) {
#if !(defined(LINUX) && defined(__KERNEL__))
	if (msg >= sys_nerr) {
	    *errorMsg = strerror(msg);
	    return FAILURE;
	}
#endif /* !(defined(LINUX) && defined(__KERNEL__)) */
    } else if (module == 0xffff) {
#if !(defined(LINUX) && defined(__KERNEL__))
	msg = STATUS_MSGNUM(-status);
	if (msg >= sys_nerr) {
	    *errorMsg = strerror(msg);
	    return FAILURE;
	}
#endif /* !(defined(LINUX) && defined(__KERNEL__)) */
    } else if (module >= (sizeof(statusMsgs) / sizeof(StatusMsgs))) {
	(void) sprintf(buffer, "Invalid module 0x%x in status 0x%x",
		module, status);
	*errorMsg = buffer;
	return FAILURE;
    } else if (msg >= statusMsgs[module].size) {
	(void) sprintf(buffer, "Invalid msg 0x%x for %s module in status 0x%x",
		msg, statusMsgs[module].name, status);
	*errorMsg = buffer;
	return FAILURE;
    }
    return SUCCESS;
}

