/*
 *  The Mirage NFS Router
 *
 *  ht.c: header table
 *
 *  @COPYRIGHT@
 *
 */

#include "nfsroute.h"
#include "md5file.h"

int VinCounter = 0;
Hash_Table HandleHashTable;

/*
 * HandleInit: Initializes the handle table module
 *
 */

Status HandleInit()
{
    Hash_InitTable(&HandleHashTable, 0, sizeof(THandleKey)/4);

    return SUCCESS;
}

/*
 * HandleCreateKey
 *
 * Create HandleKey that can be used for future calls to HandleCreate or
 * HandleLookup
 *
 */

Status HandleCreateKey(
    int pin,
    int pfs,
    int sid,
    int mountChecksum,
    THandleKey *key)
{
    key->pin = pin;
    key->pfs = pfs;
    key->sid = sid;
    key->mountChecksum = mountChecksum;

    return SUCCESS;
}

/*
 * HandleCreateKeyFromVFH
 *
 * Create HandleKey that can be used for future calls to HandleCreate or
 * HandleLookup
 *
 * Similar to HandleCreate, but gets the fields from an existing VFH
 *
 */

Status HandleCreateKeyFromVFH(struct nfs_fh *vfh, THandleKey *key)
{
    int *p;

    p = (int*) vfh->data;

    p++; /* vin */
    key->pin = *(p++);
    key->pfs = *(p++);
    key->sid = *(p++);
    key->mountChecksum = *(p++);

    return SUCCESS;
}

/*
 * HandleCreateVin
 *
 * Creates a new VIN. Right now they are just allocated sequentially.
 *
 */

int HandleCreateVin()
{
    return htonl(VinCounter++);
}

/*
 * HandleCreate
 *
 * Create an entry in the handle table.
 *
 */

Status HandleCreate(THandleKey *key, int vin, struct nfs_fh *pfh,
    THandleEntry **entryOut)
{
    Hash_Entry *h;
    Boolean isNew;
    THandleEntry *entry;

    entry = (THandleEntry*) MALLOC(sizeof(THandleEntry));

    if (pfh) {
        entry->pfh = *pfh;
    }

    entry->vin = vin;

    h = Hash_CreateEntry(&HandleHashTable, (void*) key, &isNew);
    Hash_SetValue(h, entry);

    *entryOut = entry;

    // XXX if the vin that is supplied by the caller is greater than the
    // vinCounter, then bump the VinCounter up. This is necessary to fix a
    // problem in recovery. It's somewhat of a hack and will be fixed when
    // VINs are no longer sequentially allocated.
    
    if (ntohl(vin) >= VinCounter) {
        VinCounter = ntohl(vin)+1000;
        MirageDebug("XXX Increasing vincounter (VinCounter=%d, vin=%d)\n",
            VinCounter,
            ntohl(vin));
    }

    return SUCCESS;
}

/*
 * HandleLookup
 *
 * Lookup a handle in the handle table using the key.
 *
 */

Status HandleLookup(THandleKey *key, THandleEntry **entry)
{
    Hash_Entry *h;

    h = Hash_FindEntry(&HandleHashTable, (void*) key);
    if (h == NULL) {
        return FAILURE;
    }

    (*entry) = (THandleEntry*) Hash_GetValue(h);
    if (*entry == NULL) {
        return FAILURE;
    }

    return SUCCESS;
}

/*
 * HandleLookupOrCreate
 *
 * If a handle exists, then return the handleEntry
 * Otherwise,
 *     allocate a new vin
 *     create a new handle entry
 *     return the new handle entry
 *
 * Side Effects:
 *     if a PFH is supplied, then the handle entry is updated to reflect the
 *     new PFH
 * 
 */

Status HandleLookupOrCreate(THandleKey *handleKey,
    struct nfs_fh *pfh,
    THandleEntry **handleEntry)
{
     if (HandleLookup(handleKey, handleEntry) == SUCCESS) {
         if (pfh) {
             // XXX should only do this is handleEntry->pfh is empty
             (*handleEntry)->pfh = *pfh;
         }
         return SUCCESS;
     } else {
         return HandleCreate(handleKey,
                    HandleCreateVin(),
                    pfh,
                    handleEntry);
     }

     // we should never get here
     assert(FALSE);

     return FAILURE;
}

/*
 * HandleComputeHash
 *
 * Compute the HVC checksum for the handle.
 *
 */

Status HandleComputeHash(
        char *vfh,
        char *dest)
{
    md5file context;
    char tmp[32];

    strcpy(tmp, HANDLE_HASH_SECRET);

    md5file_clear(&context);
    md5file_addn(&context, tmp, strlen(tmp)); // hash the secret
    md5file_addn(&context, vfh, HANDLE_HASH_OFFSET); // hash the handle
    md5file_hash(&context, tmp);

    memcpy(dest, tmp, HANDLE_HASH_SIZE);

    return SUCCESS;
}

/*
 * HandleCreateVFH
 *
 * Given a valid Key and a valid HandleEntry, create a VFH.
 *
 */

Status HandleCreateVFH(
        THandleKey *key,
        THandleEntry *entry,
        int pvin,
        int dosId,
        struct nfs_fh *vfh)
{
    int *p;

    memset(vfh, 0, sizeof(struct nfs_fh));

    vfh->size = NFS2_FHSIZE;

    p = (int*) vfh->data;

    *(p++) = entry->vin;
    *(p++) = key->pin;
    *(p++) = key->pfs;
    *(p++) = key->sid;
    *(p++) = key->mountChecksum;
    *(p++) = dosId;

    // make sure HANDLE_HASH_OFFSET points to where the hash is to
    // be put in the handle
    assert(HANDLE_HASH_OFFSET == (((char*) p) - ((char*) vfh->data)));

    HandleComputeHash(vfh->data, vfh->data + HANDLE_HASH_OFFSET);

    return SUCCESS;
}

/*
 * HandleValidate
 *
 * Given a VFH, check to see that the HVC checksum is valid
 *
 */

Status HandleValidate(char *vfh)
{
    char tmp[32];

    HandleComputeHash(vfh, tmp);

    if (memcmp(vfh + HANDLE_HASH_OFFSET, tmp, HANDLE_HASH_SIZE) != 0) {
        return FAILURE;
    } else {
        return SUCCESS;
    }
}

/*
 * HandleGetVINFromVFH
 *
 * Extract the VIN from the VFH.
 *
 */

int HandleGetVINFromVFH(struct nfs_fh *vfh)
{
    int *p = (int*) vfh->data;

    return *p;
}

