/*
 * ioctl.c --
 *
 *      Code for handling Mirage Ioctl's. The Ioctl is a mirage-specific
 *      extension to NFS and is used mainly for debugging (i.e. turning
 *      features on and off remotely, dumping statistics, etc)
 *
 * @COPYRIGHT@
 */


#include "nfsroute.h"

#ifdef MIRAGE_SCOUT
void MirageSilkDumpStats();
#endif

/*
 * IoctlRequest
 *
 * From a rpcHeader, determine if it is an Ioctl Request and respond
 * appropriately.
 *
 * Side Effects:
 *     features turned on or off
 *     statistics dumped
 *
 */


int IoctlRequest(char *rpcHeader,
    struct sockaddr_in *clientAddr,
    SendPacketFunc *sendFunc,
    void *sendFuncClientData)
{
    char *nfsHeader;
    char *firstAuth;
    int firstAuthSize;
    char *secondAuth;
    int secondAuthSize;
    int procNum;
    char replyBuffer[MAX_REJECT_REPLY_SIZE];
    int replySize;
    char *auth, *nfsHeaderAddr;
    int which, value, oldValue=0;

    procNum = ntohl(*RPC_PROCADDR(rpcHeader));
    if (procNum != MIRAGEPROC_IOCTL) {
        return FALSE;
    }

    firstAuth = RPC_FIRSTAUTH_ADDR(rpcHeader);
    firstAuthSize = ntohl(*OPAQUEAUTH_SIZE_ADDR(firstAuth));
    secondAuth = OPAQUEAUTH_SKIP(firstAuth, firstAuthSize);
    secondAuthSize = ntohl(*OPAQUEAUTH_SIZE_ADDR(secondAuth));
    nfsHeader = OPAQUEAUTH_SKIP(secondAuth, secondAuthSize);

    which = ntohl(*NFS2_IOCTL_WHICH_ADDR(nfsHeader));
    value = ntohl(*NFS2_IOCTL_VALUE_ADDR(nfsHeader));

    MirageDebug("  responding to ioctl\n");

    *RPC_XIDADDR(replyBuffer) = *RPC_XIDADDR(rpcHeader);
    *RPC_DIRECTION_ADDR(replyBuffer) = htonl(1);
    *RPC_REPLY_STATUS_ADDR(replyBuffer) = 0;

    // fill in an auth_none
    auth = RPC_REPLY_FIRSTAUTH_ADDR(replyBuffer);
    *OPAQUEAUTH_FLAVOR_ADDR(auth) = AUTH_NONE;
    *OPAQUEAUTH_SIZE_ADDR(auth) = 0;

    nfsHeaderAddr = OPAQUEAUTH_SKIP(auth,ntohl(*OPAQUEAUTH_SIZE_ADDR(auth)))+4;
    *((int*) (nfsHeaderAddr-4)) = 0; // XXX figure out what this is!

    switch (which) {
        case MIRAGE_IOCTL_DOS_EXTEND_ERROR:
            if (value != DosExtendErrorCodes) {
                MiragePrint("setting DosExtendErrorCodes to %d\n", value);
            }
            oldValue = DosExtendErrorCodes;
            DosExtendErrorCodes = value;
            break;

        case MIRAGE_IOCTL_DOS_RESET:
            RejectInit();
            break;

        case MIRAGE_IOCTL_DOS_ENABLE:
            if (value != DosEnable) {
                MiragePrint("setting DosEnable to %d\n", value);
            }
            oldValue = DosEnable;
            DosEnable = value;
            break;

        case MIRAGE_IOCTL_DISABLE_QUEUEDELAY:
            if (value != DisableQueueDelay) {
                MiragePrint("setting DisableQueueDelay to %d\n", value);
            }
            oldValue = DisableQueueDelay;
            DisableQueueDelay = value;
            break;

        case MIRAGE_IOCTL_DISABLE_VALIDATE:
            if (value != DisableValidate) {
                MiragePrint("setting DisableValidate to %d\n", value);
            }
            oldValue = DisableValidate;
            DisableValidate = value;
            break;

        case MIRAGE_IOCTL_DISABLE_SCHEDULER:
            if (value != DisableScheduler) {
                MiragePrint("setting DisableScheduler to %d\n", value);
            }
            oldValue = DisableScheduler;
            DisableScheduler = value;
            break;

        case MIRAGE_IOCTL_DUMP_STATS:
#ifdef MIRAGE_SCOUT
            MirageSilkDumpStats();
            oldValue = 0;
#endif
            break;

        case MIRAGE_IOCTL_SET_NFS_PORT:
            if (value != Servers[0].nfsPort) {
                MiragePrint("setting Servers[0].nfsPort to %d\n", value);
            }
            oldValue = Servers[0].nfsPort;
            Servers[0].nfsPort = value;
            Servers[0].nfsAddr.sin_port = htons(value);
            break;

        case MIRAGE_IOCTL_DISABLE_RECOVERY:
            if (value != DisableRecovery) {
                MiragePrint("setting DisableRecovery to %d\n", value);
            }
            oldValue = DisableRecovery;
            DisableRecovery = value;
            break;
    }

    *NFS2_REPLY_STATUS_ADDR(nfsHeaderAddr) = htonl(oldValue);
    replySize = (nfsHeaderAddr-replyBuffer)+4;

    MirageDebug( "  sending ioctl-reply to %s:%d\n",
            my_inet_ntoa(clientAddr->sin_addr.s_addr),
            clientAddr->sin_port);

    sendFunc(sendFuncClientData, replyBuffer, replySize, clientAddr, NULL,
        PATH_CLIENT_NFS);

    return TRUE;
}


