/*
 * mirage-redir-user.c --
 *
 *      This is a user-level redirector/forwarder. It accepts UDP packets and
 *      fowards them to a NFS server. Similarly, it forwards the requests back
 *      to the client. It only handles one client at a time.
 *
 *      This is used to test the overhead of user-level forwarding without
 *      using any mirage functionality.
 *
 * @COPYRIGHT@
 */

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <getopt.h>
#include <rpc/rpc.h>
#include <nfs_rpcgen.h>

#define NO_TRUEFALSE
#include "nfsroute.h"


#include "mirage-config.h"
//#include "recovery-user.h"

#define SERVER_ADDRESS NFS_SERVER2_ADDR
#define SERVER_MOUNT_PORT NFS_SERVER2_MOUNTPORT
#define SERVER_NFS_PORT NFS_SERVER2_NFSPORT

int MirageLogLevel = MIRAGE_DEBUG;

int mountSocket;
int mountOutSocket;
struct sockaddr_in mountAddr;
struct sockaddr_in mountOutAddr;

/* Socket variables for the nfs protocol */

int nfsSocket;
int nfsOutSocket;
struct sockaddr_in nfsAddr;
struct sockaddr_in nfsOutAddr;

/* other addrs */

struct sockaddr_in lastNfsClientAddr, lastMountClientAddr;
struct sockaddr_in serverNfsAddr, serverMountAddr;

static struct option long_options[] =  {
    {"help", 0, 0, 'h'},
    {"verbose", 0, 0, 'v'},
    {"quiet", 0, 0, 'q'},
    {0, 0, 0, 0}};

void
ShowHelp(int argc, char **argv)
{
    struct option *cur;
    char name[1024];

    fprintf (stderr, "Usage: %s [options]\n",
             argv[0]);
    fprintf (stderr, "\noptions may include: \n");

    cur = long_options;
    while (cur->name) {
        sprintf(name,"--%s",cur->name);
        fprintf(stderr,"  %-10s -%c ", name, cur->val);
        switch (cur->has_arg) {
           default:
               break;
           case 1:
               fprintf(stderr," <argument>");
               break;
           case 2:
               fprintf(stderr," [argument]");
               break;
        }
        fprintf(stderr,"\n");
        cur++;
    }
}

void
ProcessOptions(int argc, char **argv)
{
    char c;

    while ((c = getopt_long(argc, argv, "vhq", long_options, NULL)) != EOF) {
        switch (c) {
          case 'v':
              MirageLogLevel = MIRAGE_DEBUG;
              break;
          case 'q':
              MirageLogLevel = MIRAGE_QUIET;
              break;
          case 'h':
              ShowHelp(argc,argv);
              exit(1);
        }
    }
}


/*
 * InitSockets
 *
 *    Create the mount and nfs sockets that we will use for listening and
 *    sending.
 */

void InitSockets()
{
    mountSocket = socket(AF_INET, SOCK_DGRAM, 0);

    assert(mountSocket != -1);

    memset(&mountAddr, 0, sizeof(mountAddr));
    mountAddr.sin_family = AF_INET;
    mountAddr.sin_addr.s_addr = htonl(INADDR_ANY);
    mountAddr.sin_port = htons(IN_MOUNT_PORT);

    assert(bind(mountSocket, (struct sockaddr*) &mountAddr,
            sizeof(mountAddr)) == 0);

    // for sending to server

    mountOutSocket = socket(AF_INET, SOCK_DGRAM, 0);

    assert(mountOutSocket != -1);

    memset(&mountOutAddr, 0, sizeof(mountOutAddr));
    mountOutAddr.sin_family = AF_INET;
    mountOutAddr.sin_addr.s_addr = htonl(INADDR_ANY);
    mountOutAddr.sin_port = 0;

    assert(bind(mountOutSocket, (struct sockaddr*) &mountOutAddr,
            sizeof(mountOutAddr)) == 0);

    // from clients

    nfsSocket = socket(AF_INET, SOCK_DGRAM, 0);

    assert(nfsSocket != -1);

    memset(&nfsAddr, 0, sizeof(nfsAddr));
    nfsAddr.sin_family = AF_INET;
    nfsAddr.sin_addr.s_addr = htonl(INADDR_ANY);
    nfsAddr.sin_port = htons(IN_NFS_PORT);

    assert(bind(nfsSocket, (struct sockaddr*) &nfsAddr,
            sizeof(nfsAddr)) == 0);

    // for sending to server

    nfsOutSocket = socket(AF_INET, SOCK_DGRAM, 0);

    assert(nfsOutSocket != -1);

    memset(&nfsOutAddr, 0, sizeof(nfsOutAddr));
    nfsOutAddr.sin_family = AF_INET;
    nfsOutAddr.sin_addr.s_addr = htonl(INADDR_ANY);
    nfsOutAddr.sin_port = 0;

    assert(bind(nfsOutSocket, (struct sockaddr*) &nfsOutAddr,
            sizeof(nfsOutAddr)) == 0);

    serverMountAddr.sin_family = AF_INET;
    serverMountAddr.sin_addr.s_addr = inet_addr(SERVER_ADDRESS);
    serverMountAddr.sin_port = htons(SERVER_MOUNT_PORT == -1 ?
            pmap_getport(&serverMountAddr, MOUNT_PROGRAM, MOUNTVERS, 17) :
            SERVER_MOUNT_PORT);

    serverNfsAddr.sin_family = AF_INET;
    serverNfsAddr.sin_addr.s_addr = inet_addr(SERVER_ADDRESS);
    serverNfsAddr.sin_port = htons(SERVER_NFS_PORT == -1 ?
            pmap_getport(&serverNfsAddr, NFS_PROGRAM, NFS_VERSION, 17) :
            SERVER_NFS_PORT);
}



void HandleMountRequestSocket()
{
    char buffer[16384];
    int n;
    struct sockaddr_in fromAddr;
    int fromAddrLen;

    MirageDebug("handling mount request\n");

    fromAddrLen = sizeof(fromAddr);
    n = recvfrom(mountSocket, buffer, sizeof(buffer), 0,
            (struct sockaddr*) &fromAddr, &fromAddrLen);

    if (n<=0) {
        return;
    }

    lastMountClientAddr = fromAddr;
    sendto(mountOutSocket, buffer, n, 0,
            (struct sockaddr*) &serverMountAddr, sizeof(struct sockaddr));
}

void HandleMountReplySocket()
{
    char buffer[16384];
    int n;
    struct sockaddr_in fromAddr;
    int fromAddrLen;

    MirageDebug("handling mount reply\n");

    fromAddrLen = sizeof(fromAddr);
    n = recvfrom(mountOutSocket, buffer, sizeof(buffer), 0,
            (struct sockaddr*) &fromAddr, &fromAddrLen);

    if (n<=0) {
        return;
    }

    sendto(mountSocket, buffer, n, 0,
            (struct sockaddr*) &lastMountClientAddr, sizeof(struct sockaddr));

}

void HandleNfsRequestSocket()
{
    char buffer[16384];
    int n;
    struct sockaddr_in fromAddr;
    int fromAddrLen;

    MirageDebug("handling nfs request\n");

    fromAddrLen = sizeof(fromAddr);
    n = recvfrom(nfsSocket, buffer, sizeof(buffer), 0,
            (struct sockaddr*) &fromAddr, &fromAddrLen);

    if (n<=0) {
        return;
    }

    lastNfsClientAddr = fromAddr;
    sendto(nfsOutSocket, buffer, n, 0,
            (struct sockaddr*) &serverNfsAddr, sizeof(struct sockaddr));
}

void HandleNfsReplySocket()
{
    char buffer[16384];
    int rpcHeaderLen;                   /* length of rpc message */
    struct sockaddr_in fromAddr;
    int fromAddrLen;

    MirageDebug("handling nfs reply\n");

    fromAddrLen = sizeof(fromAddr);
    rpcHeaderLen = recvfrom(nfsOutSocket, buffer, sizeof(buffer), 0,
            (struct sockaddr*) &fromAddr, &fromAddrLen);

    if (rpcHeaderLen<=0) {
        return;
    }

    sendto(nfsSocket, buffer, rpcHeaderLen, 0,
        (struct sockaddr*) &lastNfsClientAddr, sizeof(struct sockaddr));
}

/*
 * Process
 *
 *     Wait for datagrams to appear on the sockets and call the appropriate
 *     handler.
 *
 * Side Effects:
 *     Datagrams received and sent. Headers rewritten.
 *
 */

void Process()
{
    fd_set readfds;
    int highSocket;
    int count;

    while (TRUE) {
        FD_ZERO(&readfds);
        FD_SET(mountSocket, &readfds);
        FD_SET(mountOutSocket, &readfds);
        FD_SET(nfsSocket, &readfds);
        FD_SET(nfsOutSocket, &readfds);

        highSocket = MAX(nfsOutSocket,MAX(nfsSocket,MAX(mountOutSocket,
                mountSocket)))+1;

        MirageDebug("waiting for socket\n");
        count=select(highSocket,&readfds,NULL,NULL,0);

        if (count > 0) {
            if (FD_ISSET(mountSocket, &readfds)) {
                HandleMountRequestSocket();
            }
            if (FD_ISSET(mountOutSocket, &readfds)) {
                HandleMountReplySocket();
            }
            if (FD_ISSET(nfsSocket, &readfds)) {
                HandleNfsRequestSocket();
            }
            if (FD_ISSET(nfsOutSocket, &readfds)) {
                HandleNfsReplySocket();
            }
        }
    }
}

int main(int argc, char **argv)
{
    ProcessOptions(argc, argv);

    /* Create sockets for sending and receiving NFS messages */
    InitSockets();

    /* Wait for and handled NFS messages */
    Process();

    return 0;
}
