/*
 * mirage-user.c --
 *
 *      User-level mirage main file. Contains the support necessary for user
 *      level mirage to send and receive packets, and process options.
 *
 * @COPYRIGHT@
 */

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <getopt.h>
#include <signal.h>
#include <unistd.h>
#include <fcntl.h>

#include "nfsroute.h"
#include "nfs_rpcgen_defs.h"
#include "mirage-config.h"
#include "recovery-user.h"
#include "tt.h"

#define DOS_SCHEDULER

typedef unsigned long long uint64;

// command line parameters
int MirageLogLevel = MIRAGE_DEBUG;
int UsePortmap = FALSE;
int OutMountPort = OUT_MOUNT_PORT;
int OutNfsPort = OUT_NFS_PORT;
int CmdServerCount = 0;
int DoDetach = FALSE;

/* Socket variables for the mount protocol */

TNfsPartner partnerStatic, *UserPartner = &partnerStatic;
TMirageUserState ClientDataNfs, ClientDataNfsOut, ClientDataMount, ClientDataMountOut;

int mountSocket;
int mountOutSocket;
struct sockaddr_in mountAddr;
struct sockaddr_in mountOutAddr;

int WantToTerminate = FALSE;

uint64 WakeTime = 0;

static int tStart, tStop;

/* Socket variables for the nfs protocol */

int nfsSocket;
int nfsOutSocket;
struct sockaddr_in nfsAddr;
struct sockaddr_in nfsOutAddr;

static struct option long_options[] =  {
    {"help", 0, 0, 'h'},
    {"verbose", 0, 0, 'v'},
    {"quiet", 0, 0, 'q'},
    {"server", 1, 0, 's'},   // syntax is "--server address:mountPort:nfsPort"
    {"portmap", 0, 0, 'p'},
    {"outmountport", 1, 0, 'x'},
    {"outnfsport", 1, 0, 'y'},
    {"secure", 0, 0, 'z'},
    {"detach", 0, 0, 't'},
    {"dosdisable", 0, 0, 'd'},

    // {"noioctl", 0, 0, '1'},
    {"novalidate", 0, 0, '2'},
    {"noqueuedelay", 0, 0, '3'},
    {"noscheduler", 0, 0, '4'},
    {0, 0, 0, 0}};

/*
 * AddServerOption
 *
 * Add a new server to mirage.
 *
 * syntax is "address:mountPort:nfsPort"
 *
 * If no port numbers are given, then portMap will
 * be used to find the appropriate ports.
 *
 */

void AddServerOption(char *str)
{
    char *mountPortStr = NULL, *nfsPortStr = NULL;

    assert(str != NULL);

    mountPortStr = strchr(str,':');
    if (mountPortStr) {
        *mountPortStr = 0;
        mountPortStr++;

        nfsPortStr = strchr(mountPortStr,':');
        if (nfsPortStr) {
            *nfsPortStr = 0;
            nfsPortStr++;
        }
    }

    if (*str) {
        AddServer(str,
                (mountPortStr) ? atoi(mountPortStr) : -1,
                (nfsPortStr) ? atoi(nfsPortStr) : -1,
                UserPartner->portMapFunc, NULL);
    }
}

/*
 * ShowHelp
 *
 * displays the command line options to stdout
 *
 */

void
ShowHelp(int argc, char **argv)
{
    struct option *cur;
    char name[1024];

    fprintf (stderr, "Usage: %s [options]\n",
             argv[0]);
    fprintf (stderr, "\noptions may include: \n");

    cur = long_options;
    while (cur->name) {
        sprintf(name,"--%s",cur->name);
        fprintf(stderr,"  %-10s -%c ", name, cur->val);
        switch (cur->has_arg) {
           default:
               break;
           case 1:
               fprintf(stderr," <argument>");
               break;
           case 2:
               fprintf(stderr," [argument]");
               break;
        }
        fprintf(stderr,"\n");
        cur++;
    }
}

/*
 * ProcessOptions
 *
 * Processes command line options.
 *
 */

void
ProcessOptions(int argc, char **argv)
{
    char c;

    while ((c = getopt_long(argc, argv, "vhqs:px:y:zd1234", long_options, NULL)) != EOF) {
        switch (c) {
          case 'v':
              MirageLogLevel = MIRAGE_DEBUG;
              break;
          case 'q':
              MirageLogLevel = MIRAGE_QUIET;
              break;
          case 'h':
              ShowHelp(argc,argv);
              exit(1);
              break;
          case 's':
              CmdServerCount++;
              AddServerOption(optarg);
              break;
          case 'p':
              UsePortmap = TRUE;
              break;
          case 'x':
              OutMountPort = atoi(optarg);
              break;
          case 'y':
              OutNfsPort = atoi(optarg);
              break;
          case 'z':
              SecureOnly = TRUE;
              break;
          case 't':
              DoDetach = TRUE;
              break;
          case 'd':
              DosEnable = FALSE;
              break;

          case '2':
              DisableValidate = TRUE;
              break;
          case '3':
              DisableQueueDelay = TRUE;
              break;
          case '4':
              DisableScheduler = TRUE;
              break;
        }
    }
}


/*
 * InitSockets
 *
 *    Create the mount and nfs sockets that we will use for listening and
 *    sending.
 */

void InitSockets()
{
    mountSocket = socket(AF_INET, SOCK_DGRAM, 0);

    assert(mountSocket != -1);

    memset(&mountAddr, 0, sizeof(mountAddr));
    mountAddr.sin_family = AF_INET;
    mountAddr.sin_addr.s_addr = htonl(INADDR_ANY);
    mountAddr.sin_port = htons(IN_MOUNT_PORT);

    assert(bind(mountSocket, (struct sockaddr*) &mountAddr,
            sizeof(mountAddr)) == 0);

    // for sending to server

    mountOutSocket = socket(AF_INET, SOCK_DGRAM, 0);

    assert(mountOutSocket != -1);

    memset(&mountOutAddr, 0, sizeof(mountOutAddr));
    mountOutAddr.sin_family = AF_INET;
    mountOutAddr.sin_addr.s_addr = htonl(INADDR_ANY);
    mountOutAddr.sin_port = htons(OutMountPort);

    assert(bind(mountOutSocket, (struct sockaddr*) &mountOutAddr,
            sizeof(mountOutAddr)) == 0);

    // from clients

    nfsSocket = socket(AF_INET, SOCK_DGRAM, 0);

    assert(nfsSocket != -1);

    memset(&nfsAddr, 0, sizeof(nfsAddr));
    nfsAddr.sin_family = AF_INET;
    nfsAddr.sin_addr.s_addr = htonl(INADDR_ANY);
    nfsAddr.sin_port = htons(IN_NFS_PORT);

    assert(bind(nfsSocket, (struct sockaddr*) &nfsAddr,
            sizeof(nfsAddr)) == 0);

    // for sending to server

    nfsOutSocket = socket(AF_INET, SOCK_DGRAM, 0);

    assert(nfsOutSocket != -1);

    memset(&nfsOutAddr, 0, sizeof(nfsOutAddr));
    nfsOutAddr.sin_family = AF_INET;
    nfsOutAddr.sin_addr.s_addr = htonl(INADDR_ANY);
    nfsOutAddr.sin_port = htons(OutNfsPort);

    ClientDataNfs.socket = nfsSocket;
    ClientDataNfsOut.socket = nfsOutSocket;
    ClientDataMount.socket = mountSocket;
    ClientDataMountOut.socket = mountOutSocket;

    assert(bind(nfsOutSocket, (struct sockaddr*) &nfsOutAddr,
            sizeof(nfsOutAddr)) == 0);

    // if no servers were added, then add the defaults
/*    if (CmdServerCount == 0) {
        MirageDebug("No command-line servers listed; Adding default servers.\n");

        AddServer(NFS_SERVER1_ADDR, NFS_SERVER1_MOUNTPORT, NFS_SERVER1_NFSPORT,
                UserPartner->portMapFunc, NULL);
        AddServer(NFS_SERVER2_ADDR, NFS_SERVER2_MOUNTPORT, NFS_SERVER2_NFSPORT,
                UserPartner->portMapFunc, NULL);
    }*/

    if (ServerCount == 0) {
        MirageError("Please use the --server command to add one or more servers.\n");
        MirageError("No servers specified. Halting.\n");
        exit(1);
    }
}

/*
 * Unset_Portmap
 *
 * Removes portmap entries for Mirage.
 *
 * Side Effect: portmap entries are removed from linux portmap daemon
 *
 */

void Unset_Portmap()
{
    pmap_unset(MOUNT_PROGRAM, MOUNTVERS);
    pmap_unset(NFS_PROGRAM, NFS_VERSION);
}

/*
 * Set_Portmap
 *
 * Adds portmap entries for Mirage. Adding portmap entries allows other
 * computers to connect to Mirage without knowing the NFS and MOUNT
 * port numbers.
 *
 * Side Effect: portmap entries are added to the linux portmap daemon
 *
 */

void Set_Portmap()
{
    Unset_Portmap();
    pmap_set(MOUNT_PROGRAM, MOUNTVERS, IPPROTO_UDP, IN_MOUNT_PORT);
    pmap_set(NFS_PROGRAM, NFS_VERSION, IPPROTO_UDP, IN_NFS_PORT);
}

/*
 * SendPacketUser
 *
 * Send packets via UDP protocol.
 *
 */

Status SendPacketUser(
        void *clientData,
        char *packetPtr,
        int packetLen,
        struct sockaddr_in *clientAddr,
        struct TServer *server,
        int whichPath)
{
    TMirageUserState *state = (TMirageUserState*) clientData;
    int socket = (int) state->socket;
    int result;
    struct sockaddr_in *toAddr;

    tStop = GetTimeStamp();

    switch (whichPath) {
        case PATH_SERVER_NFS:
            toAddr = &server->nfsAddr;
            break;

        case PATH_SERVER_MOUNT:
            toAddr = &server->mountAddr;
            break;

        case PATH_CLIENT_NFS:
        case PATH_CLIENT_MOUNT:
            toAddr = clientAddr;
            break;
    }

    result = sendto(socket, packetPtr, packetLen, 0,
        (struct sockaddr*) toAddr,
        sizeof(struct sockaddr));

    if (result >=0) {
        return SUCCESS;
    } else {
        return FAILURE;
    }
}

/*
 * HandleMountRequestSocket
 *
 * Process an incoming packet on the mount request socket.
 *
 */

void HandleMountRequestSocket()
{
    char buffer[16384];
    int n;
    struct sockaddr_in fromAddr;
    int fromAddrLen;

    MirageDebug("handling mount request\n");

    fromAddrLen = sizeof(fromAddr);
    n = recvfrom(mountSocket, buffer, sizeof(buffer), 0,
            (struct sockaddr*) &fromAddr, &fromAddrLen);

    if (n<=0) {
        return;
    }

    HandleMountRequest(UserPartner, &ClientDataMountOut, buffer, n, &fromAddr);
}

/*
 * HandleMountReplySocket
 *
 * Process an incoming packet on the mount reply socket.
 *
 */

void HandleMountReplySocket()
{
    char buffer[16384];
    int n;
    struct sockaddr_in fromAddr;
    int fromAddrLen;

    MirageDebug("handling mount reply\n");

    fromAddrLen = sizeof(fromAddr);
    n = recvfrom(mountOutSocket, buffer, sizeof(buffer), 0,
            (struct sockaddr*) &fromAddr, &fromAddrLen);

    if (n<=0) {
        return;
    }

    tStart = GetTimeStamp();
    HandleMountReply(UserPartner, &ClientDataMount, buffer, n, &fromAddr);
}

/*
 * ProcessQueue
 *
 * Pop entries off the DoS queue and process them. Each time an entry is
 * processed, the wakeTime is updated to schedule when the next entry can
 * be processed.
 *
 */

void ProcessQueue()
{
    TSchedRequest *req;
    uint64 tNow = TimeGetUS();
    int cost = 100;

/*    if (TransactGetOutstanding() > 4) {
        TransactPurge(10);
        if (TransactGetOutstanding() > 4) {
            return;
        }
    }*/

   if ((!DisableQueueDelay) && (tNow < WakeTime)) {
        return;
    }

    if (SchedPopRequest(&req) == SUCCESS) {
        HandleNfsRequest(UserPartner, &ClientDataNfsOut, &ClientDataNfs,
                req->packetPtr, req->packetLen, &req->fromAddr);

        cost = req->schedCost;

        SchedFreeRequest(req);
    }

    WakeTime = tNow + cost/7;
}

/*
 * DestroyNfsRequest
 *
 * This is a destructor for an TSchedRequest struct. It is passed into
 * the DoS scheduler so the DoS scheduler can deallocate fields that were
 * allocated by the mirage-user module.
 *
 */

void DestroyNfsRequest(TSchedRequest *req)
{
    FREE(req->packetPtr);
}

/*
 * ScheduleNFSRequest
 *
 * Enqueue an NFS request into the DoS scheduler.
 *
 */

void ScheduleNfsRequest(char *packetPtr,
    int packetLen,
    struct sockaddr_in *fromAddr)
{
    TSchedRequest *req;

    int dosId, dosScore;
    char *schedBuffer;

    if (ValidateRpcRequest(packetPtr, packetLen, &dosId, &dosScore) !=
        SUCCESS) {
        return;
    }

    req = SchedInitRequest(dosId, dosScore, DestroyNfsRequest);

    schedBuffer = (char *) MALLOC(packetLen);
    memcpy(schedBuffer, packetPtr, packetLen);

    req->packetPtr = schedBuffer;
    req->packetLen = packetLen;
    req->fromAddr = *fromAddr;

    SchedEnqueue(req);
}

/*
 * HandleNFSRequestSocket
 *
 * Process an incoming packet on the NFS request socket.
 *
 * The request may be processed directly (if DoS is disabled) or it may
 * be enqueued into the DoS scheduler.
 *
 */

void HandleNfsRequestSocket()
{
    char buffer[16384];
    int n;
    struct sockaddr_in fromAddr;
    int fromAddrLen;

    MirageDebug("handling nfs request\n");

    fromAddrLen = sizeof(fromAddr);
    n = recvfrom(nfsSocket, buffer, sizeof(buffer), 0,
            (struct sockaddr*) &fromAddr, &fromAddrLen);

    if (n<=0) {
        return;
    }

    if (!DisableScheduler) {
        ScheduleNfsRequest(buffer, n, &fromAddr);
    } else {
        HandleNfsRequest(UserPartner, &ClientDataNfsOut, &ClientDataNfs, buffer,
            n, &fromAddr);
    }
}

/*
 * HandleNfsReplySocket
 *
 * Process an incoming packet on the nfs reply socket.
 *
 */

void HandleNfsReplySocket()
{
    char buffer[16384];
    int rpcHeaderLen;                   /* length of rpc message */
    struct sockaddr_in fromAddr;
    int fromAddrLen;

    MirageDebug("handling nfs reply\n");

    fromAddrLen = sizeof(fromAddr);
    rpcHeaderLen = recvfrom(nfsOutSocket, buffer, sizeof(buffer), 0,
            (struct sockaddr *) &fromAddr, &fromAddrLen);

    if (rpcHeaderLen<=0) {
        return;
    }

    HandleNfsReply(UserPartner, &ClientDataNfs, buffer, rpcHeaderLen, &fromAddr);
}

/*
 * SigHandler
 *
 * Signal handler used to allow graceful shutdown of Mirage via signals.
 *
 */

void SigHandler(int which)
{
    switch (which) {
        case SIGINT:
            MirageDebug("SIGINT received");
            WantToTerminate = TRUE;
            break;
        case SIGHUP:
            MirageDebug("SIGHUP received");
            WantToTerminate = TRUE;
            break;
        case SIGTERM:
            MirageDebug("SIGTERM received");
            WantToTerminate = TRUE;
            break;
    }

    signal(which, SigHandler);
}


/*
 * Process
 *
 *     Wait for datagrams to appear on the sockets and call the appropriate
 *     handler.
 *
 * Side Effects:
 *     Datagrams received and sent. Headers rewritten.
 *
 */

void Process()
{
    fd_set readfds;
    int highSocket;
    int count;
    struct timeval timeout;

    while (!WantToTerminate) {
        FD_ZERO(&readfds);
        FD_SET(mountSocket, &readfds);
        FD_SET(mountOutSocket, &readfds);
        FD_SET(nfsSocket, &readfds);
        FD_SET(nfsOutSocket, &readfds);

        highSocket = MAX(nfsOutSocket,MAX(nfsSocket,MAX(mountOutSocket,
                mountSocket)))+1;

        timeout.tv_sec = 0;
        timeout.tv_usec = ((SchedNumWaiting>0) && (!DisableScheduler)) ? 0 : 100;

        count = select(highSocket,&readfds,NULL,NULL,&timeout);

        if (count > 0) {
            tStart = GetTimeStamp();

            if (FD_ISSET(mountSocket, &readfds)) {
                HandleMountRequestSocket();
            }
            if (FD_ISSET(mountOutSocket, &readfds)) {
                HandleMountReplySocket();
            }
            if (FD_ISSET(nfsSocket, &readfds)) {
                HandleNfsRequestSocket();
            }
            if (FD_ISSET(nfsOutSocket, &readfds)) {
                HandleNfsReplySocket();
            }
        }

        if (!DisableScheduler) {
            ProcessQueue();
        }
    }
}

/*
 * Detach
 *
 * Detach Mirage from the console. Used when operating mirage as a Daemon
 *
 */

void Detach(void)
{
  int pid = fork();
  int i;

  if (pid<0) {
    perror("unable to fork!");
    exit(1);
  }

  if (pid)
    exit(0);

  setsid();

  fprintf(stderr,"\nMirage Detached form terminal.\n");

  i=open("/dev/null",2);
  dup2(i,1);
  dup2(i,2);
}

/*
 * main
 *
 * Do everything that's necessary to start up Mirage, then call Process() to
 * enter the processing loop.
 *
 */

int main(int argc, char **argv)
{
    UserPartner->sendPacketFunc = SendPacketUser;

    RecoveryPartnerInitUser(UserPartner);

    ProcessOptions(argc, argv);

    /* initialize modules */
    RejectInit();
    UmtInit();
    TransactInit();
    HandleInit();
    ScoreInit();
    StolenHandleInit();

    /* Create sockets for sending and receiving NFS messages */
    InitSockets();

    /* install ourself into the mount/nfs portmap entries */
    if (UsePortmap) {
        Set_Portmap();
    }

    signal(SIGHUP, SigHandler);

    if (DoDetach) {
        Detach();
    }

    /* Wait for and handled NFS messages */
    Process();

    if (UsePortmap) {
        Unset_Portmap();
    }

    return 0;
}
