/*
 * nfsroute.c --
 *
 *      Core of mirage.
 *
 * @COPYRIGHT@
 */

#include "nfsroute.h"

#ifdef REWRITE_INODE_NUMBER
#define REWRITE_READDIR
#endif

int DosIdCounter = 0;

TServer Servers[MAX_SERVERS];
int ServerCount = 0;
int SecureOnly = FALSE;
int DisableValidate = FALSE;
int DisableQueueDelay = FALSE;
int DisableScheduler = FALSE;
int DisableRecovery = FALSE;
int DosExtendErrorCodes = FALSE;
int DosEnable = TRUE;
int lastMountServer;

int lastProcNum;
TTransactEntry *lastTransactEntry;
TServer *lastServer;

char *NfsReqNames[NUM_NFS_REQNAMES] = {
    "Null", "GetAttr", "SetAttr", "Root",
    "Lookup", "ReadLink", "Read", "WriteCache",
    "Write", "Create", "Remove", "Rename",
    "Link", "SymLink", "Mkdir", "Rmdir",
    "ReadDir", "StatFs"};

#ifdef MIRAGE_SCOUT
static char *
strdup (const char * orig)
{
    size_t len = strlen(orig) + 1;
    char * new = heapAlloc(len);

    memcpy(new, orig, len);
    return new;
}
#endif

/*
 * my_inet_ntoa
 *
 * replacement for inet_ntoa because (for some unknown reason), the stock
 * inet_ntoa was taking far longer than necessary.
 *
 */
char *my_inet_ntoa(int inaddr)
{
    unsigned char *c = (unsigned char *) &inaddr;
    static char dest[1024];

    sprintf(dest, "%d.%d.%d.%d", *c, *(c+1), *(c+2), *(c+3));

    return dest;
}


/*
 * my_inet_addr
 *
 * replacement for inet_addr.
 *
 */

int my_inet_addr(char *addr)
{
    static char tmp[1024];
    char *tmpPtr = tmp, *start = tmp;
    int done = FALSE;
    unsigned int build  = 0;
    unsigned int count = 0;

    strcpy(tmp, addr);

    while (!done) {
        switch (*tmpPtr) {
            case 0:
                done = TRUE;
            case '.':
                *tmpPtr = 0;
                build = build + (((unsigned) atoi(start)) << (count*8));
                start = tmpPtr+1;
                count++;
                break;
        }
        tmpPtr++;
    }

    return (int) build;
}

/*
 * GenerateDosId
 *
 * Return a denial of service identifier
 *
 */

int GenerateDosId(struct sockaddr_in *fromAddr)
{
// for now, return sequential DosId's to facilitate test programs
    return (DosIdCounter++);
}

/*
 * ResolveAddress
 *
 * Given an ASCII string that represents an IP address or machine name, return
 * the IP address.
 *
 */

int ResolveAddress(char *addrStr, int *addr)
{
    *addr = my_inet_addr(addrStr);

#ifndef MIRAGE_SCOUT
    if ((*addr == INADDR_NONE) || (*addr == 0)) {
        struct hostent *hostEnt;

        hostEnt = gethostbyname(addrStr);

        if (hostEnt) {
            *addr = *(int *) hostEnt->h_addr;
        }
    }
#endif

    return ((*addr != INADDR_NONE) && (*addr != 0));
}

/*
 * HandleToString
 *
 *     Given an NFS handle, return a printable string. Used for debugging
 */

char *HandleToString(struct nfs_fh *fh)
{
    static char buf[1024];
    char *bufPtr = buf;
    int i;
    unsigned char *x = (unsigned char *) fh->data;

    for (i=0; i<fh->size; i++) {
        bufPtr += sprintf(bufPtr, "%02x", *x);
        x++;
    }

    return buf;
}

/*
 * AddServer
 *
 *     Create a new TServer structure and add it to the array of servers.
 *
 * Side Effects:
 *     Server is added to array. ServerCount is incremented.
 */

TServer *AddServer(char *address,
    int mountPort,
    int nfsPort,
    PortMapFunc *portMapFunc,
    void *clientData)
{
    TServer *newServer;

    assert(ServerCount < MAX_SERVERS);

    newServer = &Servers[ServerCount];

    memset(newServer, 0, sizeof(TServer));

    newServer->sid = ServerCount;

    newServer->address = strdup(address);
    newServer->mountPort = mountPort;
    newServer->nfsPort = nfsPort;

    memset(&newServer->nfsAddr, 0, sizeof(struct sockaddr_in));
    newServer->nfsAddr.sin_family = AF_INET;

    if (!ResolveAddress(newServer->address,
            &newServer->nfsAddr.sin_addr.s_addr)) {
        MirageError("Unable to find server %s\n", newServer->address);
        return NULL;
    }

    memset(&newServer->mountAddr, 0, sizeof(struct sockaddr_in));
    newServer->mountAddr.sin_family = AF_INET;
    newServer->mountAddr.sin_addr.s_addr = newServer->nfsAddr.sin_addr.s_addr;

    // map the nfs port, calling portmap if necessary
    if (nfsPort <= 0) {
        if (!portMapFunc) {
            MirageDebug("No portmap func for sever %s\n", address);
            return NULL;
        }
        if (portMapFunc(clientData, &newServer->nfsAddr, 100003, 2,
                &nfsPort) != SUCCESS) {
            MirageDebug("Portmap nfs func failed for server %s\n", address);
            return NULL;
        }
    }
    newServer->nfsPort = nfsPort;
    newServer->nfsAddr.sin_port = htons(newServer->nfsPort);

    // map the mount port, calling portmap if necessary
    if (mountPort <= 0) {
        if (!portMapFunc) {
            MirageDebug("No portmap func for sever %s\n", address);
            return NULL;
        }
        if (portMapFunc(clientData, &newServer->mountAddr, 100005, 1,
                &mountPort) != SUCCESS) {
            MirageDebug("Portmap mount func failed for server %s\n", address);
            return NULL;
        }
    }
    newServer->mountPort = mountPort;
    newServer->mountAddr.sin_port = htons(newServer->mountPort);

    ServerCount++;

    MirageDebug("added server %s mountPort=%d nfsPort=%d\n",
        address, mountPort, nfsPort);

    return newServer;
}

/*
 * LookupServerNumber
 *
 *    Given a server's IP address, look it up in the Server array and return
 *    it's number.
 *
 * Results:
 *    The number of the server
 */

int LookupServerNumber(struct sockaddr_in *fromAddr)
{
    int i;

    for (i=0; i<ServerCount; i++) {
        if (Servers[i].mountAddr.sin_addr.s_addr == fromAddr->sin_addr.s_addr)
            return i;
    }

    assert(FALSE);

    return -1;
}

/*
 * HandleMountRequest
 *
 *    Assumes a UDP message is waiting on mountSocket. Receive the message,
 *    rewrite it if necessary, and send it to the server(s).
 */

void HandleMountRequest(
    TNfsPartner *partner,
    void *clientData,
    char *rpcHeaderPtr,
    int packetLen,
    struct sockaddr_in *fromAddr)
{
    char *mountHeaderPtr;
    int i;
    TTransactKey transactKey;
    TTransactEntry *transactEntry;
    int uid, gid;
    int rpcHeader_xid, rpcHeader_proc;
    TUmtEntry *umtEntry;

    if ((SecureOnly) && (ntohs(fromAddr->sin_port) >= 1024)) {
        MirageError("Received request from insecure port (%d) -- dropping\n",
            ntohs(fromAddr->sin_port));
        return;
    }

    xdr_getuidgid(RPC_FIRSTAUTH_ADDR(rpcHeaderPtr), &uid, &gid);
    MirageDebug(
        "from %s:%d uid=%d gid=%d\n",
        my_inet_ntoa(fromAddr->sin_addr.s_addr),
        ntohs(fromAddr->sin_port),
        ntohl(uid),
        ntohl(gid));

    // add entry to UMT and rewrite the XID
    UmtCreate(*RPC_XIDADDR(rpcHeaderPtr),
        fromAddr->sin_addr.s_addr,
        fromAddr->sin_port,
        RPC_XIDADDR(rpcHeaderPtr),
        &umtEntry);

    mountHeaderPtr = xdr_decode_rpcrequest(rpcHeaderPtr, &rpcHeader_xid, &rpcHeader_proc);

    TransactCreateKey(*RPC_XIDADDR(rpcHeaderPtr),
        fromAddr->sin_addr.s_addr,
        &transactKey);

    TransactCreate(&transactKey,
        ntohl(*RPC_PROCADDR(rpcHeaderPtr)),
        -1,
        GetMountChecksum(ntohl(*MNT_MNTPATHLEN_ADDR(mountHeaderPtr)),
                MNT_MNTPATHDATA_ADDR(mountHeaderPtr)),
        uid,
        gid,

        GenerateDosId(fromAddr),

        &transactEntry);

/*    TransactFillInvalidDosFields(transactEntry,
        NULL, 9999, -1, NULL);*/

    lastTransactEntry = transactEntry;

    MirageDebug("rpc xid=%u proc=%u\n",
        rpcHeader_xid,
        rpcHeader_proc);

    // send the request to all of the servers. XXX should only do this for
    // mount requests.

    lastProcNum = 999;

    for (i = 0; i<ServerCount; i++) {
        transactEntry->refCount++;
        umtEntry->refCount++;
        lastServer = &Servers[i];
        partner->sendPacketFunc(clientData, rpcHeaderPtr, packetLen,
                NULL, &Servers[i], PATH_SERVER_MOUNT);
    }

    // we've overcounted the refcount by one, so decrement it
    TransactFree(transactEntry);
    UmtFree(umtEntry);
}

/*
 * RewriteMountReply
 *
 *     Buffer is a pointer to the NFS header. Rewrite the header as necessary
 */

void RewriteMountReply(
    TNfsPartner *partner,
    void *clientData,
    char *buffer,
    int server,
    TTransactEntry *transactPtr,
    int *success)
{
    struct nfs_fh fh;
    struct nfs_fh rewriteHandle;
    struct mnt_fhstatus result = {0, &fh};

    THandleKey handleKey;
    THandleEntry *handleEntry;

    xdr_decode_fhstatus(buffer, &result);

    if (result.status == 0) {
        int mount_fsid, mount_fileid;

        RecoveryGetMountAttr(
                partner,
                clientData,
                Servers[server].mountAddr,
                Servers[server].nfsAddr,
                result.fh->data,
                &mount_fsid,
                &mount_fileid);

        HandleCreateKey(htonl(mount_fileid),
            htonl(mount_fsid),
            server,
            transactPtr->mountChecksum,
            &handleKey);

        assert(HandleLookupOrCreate(&handleKey, result.fh, &handleEntry)
                == SUCCESS);

        HandleCreateVFH(&handleKey, handleEntry, 9999, transactPtr->dosId,
                &rewriteHandle);

        xdr_encode_handle(buffer+4, &rewriteHandle);
        MirageDebug( "  orig mnt %s\n",
            HandleToString(result.fh));
        MirageDebug( "  rewt mnt %s\n",
            HandleToString(&rewriteHandle));

        *success = TRUE;
    } else {
        MirageDebug("  mount returns error %d\n", result.status);
        *success = FALSE;
    }
}

/*
 * HandleMountReply
 *
 *    Assumes a UDP message is waiting on mountOutSocket. Receive the message,
 *    rewrite it if necessary, and send it to the server(s).
 */

void HandleMountReply(
    TNfsPartner *partner,
    void *clientData,
    char *rpcHeaderPtr,
    int packetLen,
    struct sockaddr_in *fromAddr)
{
    char *bufPtr;
    int rpcHeader_xid, rpcHeader_stat;
    struct sockaddr_in toAddr;
    int server;
    int success = TRUE;
    TUmtEntry *umtEntry;
    TTransactKey transactKey;
    TTransactEntry *transactEntry;

    server = LookupServerNumber(fromAddr);
    lastMountServer = server;

    MirageDebug("from %s:%d\n",
        my_inet_ntoa(fromAddr->sin_addr.s_addr),
        ntohs(fromAddr->sin_port));

    if (UmtLookup(*RPC_XIDADDR(rpcHeaderPtr), &umtEntry) != SUCCESS) {
        MirageDebug("couldn't lookup umt entry\n");
        return;
    }

    assert(TransactCreateKey(*RPC_XIDADDR(rpcHeaderPtr),
        umtEntry->clientIp,
        &transactKey) == SUCCESS);

    toAddr.sin_family = AF_INET;
    toAddr.sin_addr.s_addr = umtEntry->clientIp;
    toAddr.sin_port = umtEntry->clientPort;

    bufPtr = xdr_decode_rpcreply(rpcHeaderPtr, &rpcHeader_xid, &rpcHeader_stat);

    MirageDebug("rpc xid=%u stat=%u\n",
        rpcHeader_xid,
        rpcHeader_stat);

    if (TransactLookup(&transactKey, &transactEntry) == SUCCESS) {
        lastTransactEntry = transactEntry;

        lastProcNum = 100+transactEntry->proc;
        switch (transactEntry->proc) {
            case MNTPROC_MNT:
                RewriteMountReply(partner, clientData, bufPtr, server,
                    transactEntry,&success);
                break;
        }

        /* Only forward the reply if it is successful. XXX We should count
         * how many replies we've seen, and forward a failure if all replies
         * were failures XXX
         */

        if (success) {
            *RPC_XIDADDR(rpcHeaderPtr) = umtEntry->oldXid;

            MirageDebug("  sending reply to %s:%d\n",
                    my_inet_ntoa(toAddr.sin_addr.s_addr),
                    ntohs(toAddr.sin_port));

            partner->sendPacketFunc(clientData, rpcHeaderPtr, packetLen,
                    &toAddr, NULL, PATH_CLIENT_MOUNT);
        } else {
            MirageDebug("  dropping negetive mount reply\n");
        }

        TransactFree(transactEntry);
    }

    UmtFree(umtEntry);
}

/*
 * RecoverHandle
 *
 * Perform recovery to get a PFH from a handle key.
 *
 * Side Effects: A lot of RPC calls to the server to do a recursive directory
 * search.
 *
 */

Status RecoverHandle(TNfsPartner *partner,
    void *clientData,
    THandleKey *handleKey,
    struct nfs_fh *fh)
{
    TServer *server = &Servers[handleKey->sid];

    if (Recovery(
            partner,
            clientData,
            server->mountAddr,
            server->nfsAddr,
            handleKey->mountChecksum,
            ntohl(handleKey->pin),
            ntohl(handleKey->pfs),
            fh->data)) {

        fh->size = NFS2_FHSIZE;
        return SUCCESS;
    } else {
        return FAILURE;
    }
}

/*
 * RewriteNfsHandleRequest
 *
 *     Assumes that buffer is a pointer to the part of the NFS header that
 *     contains a handle. Translare the PH to a VH. Also returns the server
 *     number.
 *
 * Side Effects:
 *     Header is rewritten
 */

Status RewriteNfsHandleRequest(
    TNfsPartner *partner,
    void *clientData,
    char *buffer,
    int *server)
{
    struct nfs_fh fh;
    struct nfs_fh rewriteHandle;

    THandleKey handleKey;
    THandleEntry *handleEntry;

    xdr_decode_fhandle(buffer, &fh);

    MirageDebug( "  orig fh %s\n",
        HandleToString(&fh));

    HandleCreateKeyFromVFH(&fh, &handleKey);
    if (HandleLookup(&handleKey, &handleEntry) != SUCCESS) {
        struct nfs_fh pfh;
        if (RecoverHandle(partner, clientData, &handleKey, &pfh) == SUCCESS) {
            assert(HandleCreate(&handleKey,
                    HandleGetVINFromVFH(&fh),
                    &pfh,
                    &handleEntry) == SUCCESS);
        } else {
            MirageError( "Unable to recover handle\n");
            return FAILURE;
        }
    }

    rewriteHandle = handleEntry->pfh;

    xdr_encode_handle(buffer, &rewriteHandle);

    MirageDebug( "  rewt fh %s\n",
        HandleToString(&rewriteHandle));

    *server = handleKey.sid;

    return SUCCESS;
}

/*
 * ValidateNfsRequest
 *
 * Given an NFS request, verify that it is legitimate by checking the VFHs
 * match their MCHs
 *
 */

Status ValidateNfsRequest(char *nfsHeaderPtr, int rpcHeader_proc)
{
    Status result;

    if (DisableValidate) {
        return SUCCESS;
    }

    switch (rpcHeader_proc) {
        case NFSPROC_CREATE:
        case NFSPROC_GETATTR:
        case NFSPROC_SETATTR:
        case NFSPROC_LOOKUP:
        case NFSPROC_MKDIR:
        case NFSPROC_READLINK:
        case NFSPROC_READ:
        case NFSPROC_READDIR:
        case NFSPROC_REMOVE:
        case NFSPROC_RMDIR:
        case NFSPROC_STATFS:
        case NFSPROC_SYMLINK:
        case NFSPROC_WRITE:
            return HandleValidate(nfsHeaderPtr);

        case NFSPROC_RENAME:
            if ((result = HandleValidate(nfsHeaderPtr)) != SUCCESS) {
                return result;
            }
            return HandleValidate(NFS2_RENAMEARGS_TO_ADDR(nfsHeaderPtr));

        case NFSPROC_LINK:
            if ((result = HandleValidate(nfsHeaderPtr)) != SUCCESS) {
                return result;
            }
            return HandleValidate(nfsHeaderPtr+NFS2_FHSIZE);
            break;

        case MIRAGEPROC_IOCTL:
        case NFSPROC_NULL:
            break;
    }

    return SUCCESS;
}

/*
 * SendErrorReply
 *
 * Send an error reply with "errorCode" to the client.
 *
 */

int SendErrorReply(
    int xid,
    struct sockaddr_in *clientAddr,
    SendPacketFunc *sendFunc,
    void *sendFuncClientData,
    int errorCode)
{
    char replyBuffer[MAX_REJECT_REPLY_SIZE];
    int replySize;
    char *auth, *nfsHeaderAddr;

    *RPC_XIDADDR(replyBuffer) = xid;
    *RPC_DIRECTION_ADDR(replyBuffer) = htonl(1);
    *RPC_REPLY_STATUS_ADDR(replyBuffer) = 0;

    // fill in an auth_none
    auth = RPC_REPLY_FIRSTAUTH_ADDR(replyBuffer);
    *OPAQUEAUTH_FLAVOR_ADDR(auth) = AUTH_NONE;
    *OPAQUEAUTH_SIZE_ADDR(auth) = 0;

    nfsHeaderAddr = OPAQUEAUTH_SKIP(auth,ntohl(*OPAQUEAUTH_SIZE_ADDR(auth)))+4;
    *((int*) (nfsHeaderAddr-4)) = 0; // XXX figure out what this is!

    *NFS2_REPLY_STATUS_ADDR(nfsHeaderAddr) = errorCode;

    replySize = (nfsHeaderAddr-replyBuffer)+4;

    MirageDebug("  sending error-reply #%d to %s:%d\n",
            errorCode,
            my_inet_ntoa(clientAddr->sin_addr.s_addr),
            clientAddr->sin_port);

    sendFunc(sendFuncClientData, replyBuffer, replySize, clientAddr,
         NULL, PATH_CLIENT_NFS);

    return TRUE;
}

/*
 * HandleNfsRequest
 *
 *    Assumes a UDP message is waiting on nfsSocket. Receive the message,
 *    rewrite it if necessary, and send it to the server(s).
 */

void HandleNfsRequest(
    TNfsPartner *partner,
    void *clientData,
    void *dosClientData,
    char *rpcHeaderPtr,
    int packetLen,
    struct sockaddr_in *fromAddr)
{
    char *nfsHeaderPtr;
    int server;
    int rpcHeader_proc, rpcHeader_xid;
    int uid, gid;
    int dosId;
    int origXid;
    TTransactEntry *transactEntry;
    TTransactKey transactKey;

    if ((SecureOnly) && (ntohs(fromAddr->sin_port) >= 1024)) {
        MirageError("Received request from insecure port (%d) -- dropping\n",
            ntohs(fromAddr->sin_port));
        return;
    }

    nfsHeaderPtr = xdr_decode_rpcrequest(rpcHeaderPtr, &rpcHeader_xid, &rpcHeader_proc);

    // if the scheduler is disabled, then do the request validation here
    // instead.
    if ((DisableScheduler) && (!DisableValidate)) {
        if (ValidateNfsRequest(nfsHeaderPtr, rpcHeader_proc) != SUCCESS) {
            MirageDebug("header contains invalid handle(s)\n");
            return;
        }
    }

    xdr_getuidgid(RPC_FIRSTAUTH_ADDR(rpcHeaderPtr), &uid, &gid);

    dosId = *NFS2_REQUEST_DOSID_ADDR(nfsHeaderPtr);

    MirageDebug("from %s:%d uid=%d gid=%d dosid=%d\n",
        my_inet_ntoa(fromAddr->sin_addr.s_addr),
        ntohs(fromAddr->sin_port),
        ntohl(uid),
        ntohl(gid),
        dosId);

    if (IoctlRequest(rpcHeaderPtr, fromAddr, partner->sendPacketFunc, dosClientData)) {
        MirageDebug("request was handled by IoctlRequest()");
        return;
    }

    origXid = *RPC_XIDADDR(rpcHeaderPtr);

    // add entry to UMT and rewrite the XID
    UmtCreate(*RPC_XIDADDR(rpcHeaderPtr),
        fromAddr->sin_addr.s_addr,
        fromAddr->sin_port,
        RPC_XIDADDR(rpcHeaderPtr),
        NULL);

    MirageDebug("rpc xid=%u proc=%u\n",
        rpcHeader_xid,
        rpcHeader_proc);

    TransactCreateKey(*RPC_XIDADDR(rpcHeaderPtr),
        fromAddr->sin_addr.s_addr,
        &transactKey);

    TransactCreate(&transactKey,
        ntohl(*RPC_PROCADDR(rpcHeaderPtr)),
        *NFS2_REQUEST_VIN_ADDR(nfsHeaderPtr),
        *NFS2_REQUEST_MOUNTCHECKSUM_ADDR(nfsHeaderPtr),
        uid,
        gid,
        *NFS2_REQUEST_DOSID_ADDR(nfsHeaderPtr),
        &transactEntry);

    lastTransactEntry = transactEntry;

    lastProcNum = rpcHeader_proc;
    switch (rpcHeader_proc) {
        case NFSPROC_CREATE:
        case NFSPROC_GETATTR:
        case NFSPROC_SETATTR:
        case NFSPROC_LOOKUP:
        case NFSPROC_MKDIR:
        case NFSPROC_READLINK:
        case NFSPROC_READ:
        case NFSPROC_READDIR:
        case NFSPROC_REMOVE:
        case NFSPROC_RMDIR:
        case NFSPROC_STATFS:
        case NFSPROC_SYMLINK:
        case NFSPROC_WRITE:
            if (RewriteNfsHandleRequest(partner, clientData, nfsHeaderPtr, &server) != SUCCESS) {
                SendErrorReply(origXid, fromAddr, partner->sendPacketFunc,
                        dosClientData, NFSERR_STALE);
                return;
            }
            break;

        case NFSPROC_RENAME:
            if (RewriteNfsHandleRequest(partner, clientData, nfsHeaderPtr, &server) != SUCCESS) {
                SendErrorReply(origXid, fromAddr, partner->sendPacketFunc,
                        dosClientData, NFSERR_STALE);
                return;
            }
            if (RewriteNfsHandleRequest(partner, clientData, NFS2_RENAMEARGS_TO_ADDR(nfsHeaderPtr),
                    &server) != SUCCESS) {
                SendErrorReply(origXid, fromAddr, partner->sendPacketFunc,
                        dosClientData, NFSERR_STALE);
                return;
            }
            break;

        case NFSPROC_LINK:
            if (RewriteNfsHandleRequest(partner, clientData, nfsHeaderPtr, &server) != SUCCESS) {
                SendErrorReply(origXid, fromAddr, partner->sendPacketFunc,
                        dosClientData, NFSERR_STALE);
                return;
            }
            if (RewriteNfsHandleRequest(partner, clientData, nfsHeaderPtr + NFS2_FHSIZE, &server) !=
                    SUCCESS) {
                SendErrorReply(origXid, fromAddr, partner->sendPacketFunc,
                        dosClientData, NFSERR_STALE);
                return;
            }
            break;

        case NFSPROC_NULL:
            // XXX this is a hack because there is no server id in a NULL
            // request, so we use the server that the client most recently
            // mounted.
            server = lastMountServer;
            break;
    }

    lastServer = &Servers[server];

    partner->sendPacketFunc(clientData, rpcHeaderPtr, packetLen,
            NULL, &Servers[server], PATH_SERVER_NFS);
}

/*
 * RewriteNfsAttrStatRely
 *
 * Side Effects:
 *    Header is rewritten.
 */

void RewriteNfsAttrStatReply(char *buffer,
    int server,
    TTransactEntry *transactPtr)
{
    int status;
    THandleKey handleKey;
    THandleEntry *handleEntry;

    buffer = xdr_decode_nfsstat(buffer, &status);

    if (status == 0) {
        HandleCreateKey(*NFS2_ATTRSTAT_FILEID_ADDR(buffer),
            *NFS2_ATTRSTAT_FSID_ADDR(buffer),
            server,
            transactPtr->mountChecksum,
            &handleKey);

        assert(HandleLookupOrCreate(&handleKey, NULL, &handleEntry)
                == SUCCESS);

        // rewrite the fileid
#ifdef REWRITE_INODE_NUMBER
        *NFS2_ATTRSTAT_FILEID_ADDR(buffer) = handleEntry->vin;
#endif
    }
}

/*
 * RewriteNfsDiropRely
 *
 *    Assumes buffer points to a dirop_reply part of a nfs header. Translate
 *    the PH to a VH and rewrite the header.
 *
 * Side Effects:
 *    Header is rewritten.
 */

void RewriteNfsDiropReply(char *buffer,
    int server,
    TTransactEntry *transactPtr)
{
    struct nfs_fh fh;
    struct nfs_fh rewriteHandle;
    int status;
    THandleKey handleKey;
    THandleEntry *handleEntry;

    buffer = xdr_decode_nfsstat(buffer, &status);

    if (status == 0) {
        xdr_decode_fhandle(buffer, &fh);

        HandleCreateKey(*NFS2_DIROPOK_FILEID_ADDR(buffer),
            *NFS2_DIROPOK_FSID_ADDR(buffer),
            server,
            transactPtr->mountChecksum,
            &handleKey);

        assert(HandleLookupOrCreate(&handleKey, &fh, &handleEntry)
                == SUCCESS);

        // rewrite the handle
        HandleCreateVFH(&handleKey, handleEntry, transactPtr->vin,
                transactPtr->dosId,
                &rewriteHandle);

        // rewrite the fileid
#ifdef REWRITE_INODE_NUMBER
        *NFS2_DIROPOK_FILEID_ADDR(buffer) = handleEntry->vin;
#endif

        xdr_encode_handle(buffer, &rewriteHandle);
        MirageDebug( "  orig fh %s\n",
            HandleToString(&fh));
        MirageDebug( "  rewrote fh %s\n",
            HandleToString(&rewriteHandle));

    }
}

/*
 * RewriteNfsReadDirReply
 *
 * Side Effects:
 *    Inodes are rewritten
 */

void RewriteNfsReaddirReply(char *buffer,
    int server,
    TTransactEntry *transactPtr)
{
#ifdef REWRITE_READDIR
    struct nfs_fh fh;
    int status;
    THandleKey handleKey;
    THandleEntry *handleEntry;
    int *intPtr;
    int len;

    buffer = xdr_decode_nfsstat(buffer, &status);

    intPtr = (int*) buffer;
    while (*intPtr++) {
        HandleCreateKey(*intPtr,
            9999, /* XXX parents fsid should go here */
            server,
            transactPtr->mountChecksum,
            &handleKey);

        assert(HandleLookupOrCreate(&handleKey, &fh, &handleEntry)
                == SUCCESS);

        /* rewrite the inode number */
#ifdef REWRITE_INODE_NUMBER
        *intPtr = handleEntry->vin;
#endif

        /* skip over inode number */
        intPtr++;

        /* skip over name */
        len = ntohl(*(intPtr++));
        intPtr += XDR_QUADLEN(len);

        /* skip over cookie */
        intPtr++;
    }
#endif
}

/*
 * HandleNfsReply
 *
 *    Assumes a UDP message is waiting on nfsOutSocket. Receive the message,
 *    rewrite it if necessary, and send it to the server(s).
 */

Status HandleNfsReply(
    TNfsPartner *partner,
    void *clientData,
    char *rpcHeaderPtr,
    int packetLen,
    struct sockaddr_in *fromAddr)
{
    char *nfsHeaderPtr;                 /* ptr to start of nfs message */
    int rpcHeader_xid, rpcHeader_stat;
    struct sockaddr_in toAddr;
    int server;
    Status status;

    TUmtEntry *umtEntry;

    TTransactKey transactKey;
    TTransactEntry *transactEntry;

    server = LookupServerNumber(fromAddr);

    MirageDebug("from %s:%d\n",
        my_inet_ntoa(fromAddr->sin_addr.s_addr),
        ntohs(fromAddr->sin_port));

    toAddr.sin_family = AF_INET;

    // rewrite XID and lookup toAddr
    if (UmtLookup(*RPC_XIDADDR(rpcHeaderPtr), &umtEntry) != SUCCESS) {
        MirageDebug("couldn't find umt entry\n");
        return FAILURE;
    }

    assert(TransactCreateKey(*RPC_XIDADDR(rpcHeaderPtr),
        umtEntry->clientIp,
        &transactKey) == SUCCESS);

    toAddr.sin_family = AF_INET;
    toAddr.sin_addr.s_addr = umtEntry->clientIp;
    toAddr.sin_port = umtEntry->clientPort;

    nfsHeaderPtr = xdr_decode_rpcreply(rpcHeaderPtr, &rpcHeader_xid,
            &rpcHeader_stat);

    if (rpcHeader_stat == 1) {
        int rejStatus = ntohl(*RPC_REPLY_REJSTATUS_ADDR(rpcHeaderPtr));
        if (rejStatus == 1) {
            char *hvcPtr = NFS2_REQUEST_HVC_ADDR(nfsHeaderPtr);

            MirageDebug("bad rpc authentication: flagging stolen handle\n");
            StolenHandleCreate(hvcPtr, NULL);
        }
    }

    MirageDebug("rpc xid=%u stat=%u\n",
        rpcHeader_xid,
        rpcHeader_stat);

    status = TransactLookup(&transactKey, &transactEntry);
    lastTransactEntry = transactEntry;

    if (status == SUCCESS) {
        lastProcNum = transactEntry->proc;
        switch (transactEntry->proc) {
            case NFSPROC_CREATE:
            case NFSPROC_LOOKUP:
            case NFSPROC_MKDIR:
                RewriteNfsDiropReply(nfsHeaderPtr, server, transactEntry);
                break;

            case NFSPROC_READDIR:
                RewriteNfsReaddirReply(nfsHeaderPtr, server, transactEntry);
                break;

            case NFSPROC_READ:
                /* read reply is an attrstat followed by data */
            case NFSPROC_GETATTR:
            case NFSPROC_SETATTR:
            case NFSPROC_WRITE:
                RewriteNfsAttrStatReply(nfsHeaderPtr, server, transactEntry);
                break;
        }

        *RPC_XIDADDR(rpcHeaderPtr) = umtEntry->oldXid;

        MirageDebug( "sending reply to %s:%d\n",
                my_inet_ntoa(toAddr.sin_addr.s_addr),
                ntohs(toAddr.sin_port));

        partner->sendPacketFunc(clientData, rpcHeaderPtr, packetLen,
                &toAddr, NULL, PATH_CLIENT_NFS);

        TransactFree(transactEntry);
    }

    UmtFree(umtEntry);

    return SUCCESS;
}

Status ValidateRpcRequest(char *packetPtr,
    int packetLen,
    int *dosId,
    int *dosScore)
{
    char *nfsHeaderPtr;
    char *hvcPtr;
    int rpcHeader_xid, rpcHeader_proc;
    int stolenCount;

    nfsHeaderPtr = xdr_decode_rpcrequest(packetPtr,
           &rpcHeader_xid, &rpcHeader_proc);

    if (ValidateNfsRequest(nfsHeaderPtr, rpcHeader_proc) != SUCCESS) {
        MirageDebug("header contains invalid handle(s)\n");
        return FAILURE;
    }

    *dosId = *NFS2_REQUEST_DOSID_ADDR(nfsHeaderPtr);

    *dosScore = ScoreComputeNfs(rpcHeader_proc, packetLen);

    hvcPtr = NFS2_REQUEST_HVC_ADDR(nfsHeaderPtr);
    if (StolenHandleLookup(hvcPtr, &stolenCount)) {
        *dosScore = *dosScore + stolenCount*STOLEN_HANDLE_PENALTY;
    }

    return SUCCESS;
}



