/*
 * recovery-user.c --
 *
 *      User-level functions for the recovery module. These implement the
 *      low level NFS i/o that is used by recovery.c
 *
 * @COPYRIGHT@
 */

#include <sys/stat.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <assert.h>
#include "nfs_rpcgen_types.h"
#include "nfs_rpcgen_funcs.h"
#include "recovery.h"

typedef struct TRecoveryUserState {
    CLIENT *mountClient;
    CLIENT *nfsClient;
} TRecoveryUserState;

Status RecoverySetupUser(
    void *clientData,
    struct sockaddr_in mountAddr,
    struct sockaddr_in nfsAddr)
{
    TRecoveryUserState *state = (TRecoveryUserState *) clientData;
    struct sockaddr_in mountAddress, nfsAddress;
    int *mountSocket, *nfsSocket;
    struct timeval timeout;

    /* address & port of mount server */
    mountAddress = mountAddr; //.sin_family = AF_INET;
//    mountAddress.sin_addr.s_addr = inet_addr(serverAddress);
//    mountAddress.sin_port = htons(mountPort);
    mountSocket = RPC_ANYSOCK;

    /* address and port of nfs server */
    nfsAddress = nfsAddr; //.sin_family = AF_INET;
//    nfsAddress.sin_addr.s_addr = inet_addr(serverAddress);
//    nfsAddress.sin_port = htons(nfsPort);
    nfsSocket = RPC_ANYSOCK;

    /* default timeout value is set to one second */
    timeout.tv_usec = 0;
    timeout.tv_sec = 1;

    /* create client connections for mount & nfs */
    state->mountClient = clntudp_create(&mountAddress, MOUNT_PROGRAM, MOUNTVERS, timeout, &mountSocket);

    state->nfsClient = clntudp_create(&nfsAddress, NFS_PROGRAM, NFS_VERSION, timeout, &nfsSocket);
    state->nfsClient->cl_auth = authunix_create_default();

    return ((state->mountClient!=NULL) && (state->nfsClient!=NULL)) ? SUCCESS : FAILURE;
}

Status RecoveryExportUser(void *clientData,
    exports *exportsResult)
{
    TRecoveryUserState *state = (TRecoveryUserState *) clientData;
    exports *exportsResult1;

    exportsResult1 = mountproc_export_1(NULL, state->mountClient);
    if (exportsResult1) {
        *exportsResult = *exportsResult1;
    }

    return (exportsResult1) ? SUCCESS : FAILURE;
}

Status RecoveryGetAttrUser(void *clientData,
    char *fh,
    attrstat *attrStat)
{
    TRecoveryUserState *state = (TRecoveryUserState *) clientData;
    fhandleargs fhandleArgs;
    attrstat *attrStat1;

    memcpy(fhandleArgs.data, fh, FHSIZE);
    attrStat1 = nfsproc_getattr_2(&fhandleArgs, state->nfsClient);
    if (attrStat1) {
        *attrStat = *attrStat1;
    }

    return (attrStat1) ? SUCCESS : FAILURE;
}

Status RecoveryMountUser(void *clientData,
    char *mountPath,
    fhstatus *mountResult)
{
    TRecoveryUserState *state = (TRecoveryUserState *) clientData;
    dirpath mountArgs;
    fhstatus *mountResult1;

    /* mount the file system */
    mountArgs = mountPath;
    mountResult1 = mountproc_mnt_1(&mountArgs, state->mountClient);
    if (mountResult1) {
        *mountResult = *mountResult1;
    }

    return (mountResult1) ? SUCCESS : FAILURE;
}

Status RecoveryLookupUser(void *clientData,
    char *dirHandle,
    char *fn,
    diropres *lookupResult)
{
    TRecoveryUserState *state = (TRecoveryUserState *) clientData;
    diropargs lookupArgs;
    diropres *lookupResult1;

    memcpy(lookupArgs.dir, dirHandle, FHSIZE);
    lookupArgs.name = fn;
    lookupResult1 = nfsproc_lookup_2(&lookupArgs, state->nfsClient);
    if (lookupResult1) {
        *lookupResult = *lookupResult1;
    }

    return (lookupResult1) ? SUCCESS : FAILURE;
}

Status RecoveryReadDirUser(void *clientData,
    char *dirHandle,
    char *cookie,
    readdirres *readdirRes)
{
    TRecoveryUserState *state = (TRecoveryUserState *) clientData;
    readdirargs readdirArgs;
    readdirres *readdirRes1;

    memcpy(readdirArgs.dir, dirHandle, FHSIZE);

    readdirArgs.count = 1000;

    if (cookie == NULL) {
        memset(readdirArgs.cookie, 0, 4);
    } else {
        memcpy(readdirArgs.cookie, cookie, 4);
    }

    readdirRes1 = nfsproc_readdir_2(&readdirArgs, state->nfsClient);
    if (readdirRes1 != NULL) {
        *readdirRes = *readdirRes1;
    }

    return (readdirRes1) ? SUCCESS : FAILURE;
}

Status RecoveryCleanupUser(void *clientData)
{
    TRecoveryUserState *state = (TRecoveryUserState *) clientData;

    clnt_destroy(state->mountClient);
    clnt_destroy(state->nfsClient);

    return SUCCESS;
}

Status PortMapUser(
        void *clientData,
        struct sockaddr_in *serverAddr,
        int prog,
        int vers,
        int *port)
{
    *port = pmap_getport(serverAddr, prog, vers, 17);

    if (*port == 0) {
        return FAILURE;
    } else {
        return SUCCESS;
    }
}

Status RecoveryPartnerInitUser(TNfsPartner *partner)
{
    partner->portMapFunc = PortMapUser;

    partner->readDirFunc = RecoveryReadDirUser;
    partner->lookupFunc = RecoveryLookupUser;
    partner->getAttrFunc = RecoveryGetAttrUser;
    partner->mountFunc = RecoveryMountUser;
    partner->exportFunc = RecoveryExportUser;
    partner->recoverySetupFunc = RecoverySetupUser;
    partner->recoveryCleanupFunc = RecoveryCleanupUser;

    return SUCCESS;
}




