/*
 * recovery.c --
 *
 * functions for performing recovery (i.e. given a PIN/PFS, return the PFH)
 * this is done through a recursive directory search.
 *
 * This file has the high-level engine for recovery. Low-level functions
 * are in recovery-user.c (for user mode) or recovery-silk.c (for silk).
 *
 * @COPYRIGHT@
 */

#ifdef ENABLE_RECOVERY
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <assert.h>
#include <recovery.h>
#include <nfs_rpcgen.h>
#else
#include "nfsroute.h"
#endif

#define nfs_fh char

#define ENABLE_RECOVERY

int GetMountChecksum(int len, char *data)
{
    int checksum = 0;

    while (len > 0) {
        checksum = (checksum<<4) ^ *data;
        data++;
        len--;
    }

    return checksum;
}

#ifdef ENABLE_RECOVERY

int
TryFileName(TNfsPartner *partner,
    void *clientData,
    nfs_fh *dirHandle,
    char *filename,
    int fileid,
    int fsid,
    nfs_fh *handle)
{
    diropres lookupResult;
    fattr *attributes;

    MirageDebug("Trying filename %s\n", filename);

    if (partner->lookupFunc(clientData, dirHandle, filename, &lookupResult) !=
        SUCCESS) {
        return FALSE;
    }

    if (lookupResult.status != 0) {
        return FALSE;
    }

    attributes = &lookupResult.diropres_u.diropok.attributes;

    if (attributes == NULL) {
        return FALSE;
    }

    MirageDebug("    fsid = %d, fileid=%d\n", attributes->fsid, attributes->fileid);

    if ((attributes->fsid = fsid) && (attributes->fileid == fileid)) {
        memcpy(handle, lookupResult.diropres_u.diropok.file, FHSIZE);
        return TRUE;
    }

    if (((attributes->mode & 0040000) !=0 ) &&
            (strcmp(filename,".")!=0) &&
            (strcmp(filename,"..")!=0)) {
        MirageDebug( "Trying Directory %s\n", filename);
        if (TryDirectory(partner, clientData,
                lookupResult.diropres_u.diropok.file,
                fileid,
                fsid,
                handle)) {
            return TRUE;
        }
    }

    return FALSE;
}

int
TryDirectory(
    TNfsPartner *partner,
    void *clientData,
    nfs_fh *dirHandle,
    int fileid,
    int fsid,
    nfs_fh *handle)
{
    readdirres readdirResult;
    char *lastCookie;
    Status readdirStatus;

    readdirStatus = partner->readDirFunc(clientData, dirHandle, NULL, &readdirResult);
    while ((readdirStatus == SUCCESS) && (readdirResult.status == NFS_OK)) {
        struct entry *entry;

        entry = readdirResult.readdirres_u.readdirok.entries;
        while (entry != NULL) {
            if (TryFileName(partner, clientData, dirHandle, entry->name, fileid, fsid,
                     handle)) {
                return TRUE;
            }
            lastCookie = entry->cookie;
            entry = entry->nextentry;
        }

        if (readdirResult.readdirres_u.readdirok.eof) {
            readdirStatus = FAILURE;
        } else {
            readdirStatus = partner->readDirFunc(clientData, dirHandle,
                    lastCookie, &readdirResult);
        }
    }

    MirageDebug("failed to find in directory\n");

    return FALSE;
}

int
TryMountPoint(
    TNfsPartner *partner,
    void *clientData,
    char *mountpath,
    int fileid,
    int fsid,
    nfs_fh *handle)
{
    fhstatus mountResult;

    if (partner->mountFunc(clientData, mountpath, &mountResult) == FAILURE) {
        return FALSE;
    }

    if (mountResult.status != 0) {
        MirageError("mount %s failed\n", mountpath);
        return FALSE;
    }

    return TryDirectory(partner, clientData, mountResult.fhstatus_u.directory,
        fileid, fsid, handle);
}

int Recovery(TNfsPartner *partner,
    void *clientData,
    struct sockaddr_in mountAddr,
    struct sockaddr_in nfsAddr,
    int mountChecksum,
    int fileid,
    int fsid,
    char *handle)
{
    exports exportsResult;
    int found = FALSE;
    exportnode *curExport;

    if (DisableRecovery) {
        return FALSE;
    }

    MirageDebug("recovery fsid=%d fileid=%d\n", fsid, fileid);

    if (partner->recoverySetupFunc(clientData, mountAddr, nfsAddr) !=
            SUCCESS) {
        return FALSE;
    }

    if (partner->exportFunc(clientData, &exportsResult) != SUCCESS) {
        MirageError(" unable to get exports\n");
        goto done;
    }

    curExport = exportsResult;
    while (curExport != NULL) {
        if (GetMountChecksum(strlen(curExport->ex_dir),
                curExport->ex_dir) == mountChecksum) {
             MirageDebug("trying mountpoint %s\n", curExport->ex_dir);
             if (TryMountPoint(partner, clientData, curExport->ex_dir,
                     fileid, fsid, handle)) {
                 MirageDebug("found!\n");
                 found = TRUE;
                 goto done;
             }
        }
        curExport = curExport->ex_next;
    }

done:
    partner->recoveryCleanupFunc(clientData);
    return found;
}

int RecoveryGetMountAttr(
    TNfsPartner *partner,
    void *clientData,
    struct sockaddr_in mountAddr,
    struct sockaddr_in nfsAddr,
    char *fh,
    int *fsid,
    int *fileid)
{
    attrstat attrStat;
    int found = FALSE;

    *fsid = 9999;
    *fileid = 9999;

    if (DisableRecovery) {
        return FALSE;
    }

    if (partner->recoverySetupFunc(clientData, mountAddr, nfsAddr) !=
            SUCCESS) {
        return FALSE;
    }

    if (partner->getAttrFunc(clientData, fh, &attrStat) != SUCCESS) {
        found = FALSE;
        goto done;
    }

    if (attrStat.status == NFS_OK) {
        *fsid = attrStat.attrstat_u.attributes.fsid;
        *fileid = attrStat.attrstat_u.attributes.fileid;
        found = TRUE;
        goto done;
    }

done:
    partner->recoveryCleanupFunc(clientData);
    return found;
}

#else

/* XXX Stubs if recovery is disabled */

int Recovery(
    TNfsPartner *partner,
    void *clientData,
    struct sockaddr_in mountAddr,
    struct sockaddr_in nfsAddr,
    int mountChecksum,
    int fileid,
    int fsid,
    char *handle)
{
    memset(handle, 0, 32);

    return FALSE;
}

int RecoveryGetMountAttr(
    TNfsPartner *partner,
    void *clientData,
    struct sockaddr_in mountAddr,
    struct sockaddr_in nfsAddr,
    char *fh,
    int *fsid,
    int *fileid)
{
//    memset(fh, 0, 32);
    *fsid = 9999;
    *fileid = 9999;

    return FALSE;
}

#endif




