/*
 * schedule.c --
 *
 *      The DoS scheduler. NFS requests are enqueued into the DoS scheduler
 *      and then dequeued in a fair manner to enforce DoS secheduling.
 *
 *      The scheduler uses a round-robin list of clients, and each client has
 *      a circular queue of pending requests. Charge based proportional
 *      scheduling is used to process the list of clients when popping
 *      requests for service.
 *
 * @COPYRIGHT@
 */

#include "nfsroute.h"

TSchedHeader *SchedList = NULL;
TSchedHeader *RoundRobinPtr = NULL;
int SchedNumWaiting;

/*
 * CircularInit
 *
 * Initializes a circular queue
 *
 */

void CircularInit(TCircularQueue *q, int size)
{
    q->size = size;
    q->head = 0;
    q->tail = 0;
    q->count = 0;
}

/*
 * CircularFull
 *
 * Returns TRUE if a circular queue is full
 *
 */

int CircularFull(TCircularQueue *q)
{
    return (q->count == q->size);
}

/*
 * CircularEmpty
 *
 * Returns TRUE if a circular queue is empty
 *
 */

int CircularEmpty(TCircularQueue *q)
{
    return (q->count == 0);
}

/*
 * CircularEnqueue
 *
 * Enqueue an entry into a circular queue. Return FAILURE if there is no room.
 *
 */

Status CircularEnqueue(TCircularQueue *q, void **item)
{
    if (q->count == q->size) {
        return FAILURE;
    }

    q->slots[q->head] = item;
    q->head = (q->head+1) % q->size;
    q->count++;

    return SUCCESS;
}

/*
 * CircularDequeue
 *
 * Remove an item from a circular queue. Return FAILURE if it is empty.
 *
 */

Status CircularDequeue(TCircularQueue *q, int peekOnly, void **item)
{
    void *currentItem;

    if (q->count == 0) {
        return FAILURE;
    }

    currentItem = q->slots[q->tail];

    if (!peekOnly) {
        q->tail = (q->tail+1) % q->size;
        q->count--;
    }

    *item = currentItem;
    return SUCCESS;
}

/*
 * SchedAllocRequest
 *
 * Allocate a scheduler request structure
 *
 */

TSchedRequest *SchedAllocRequest()
{
    TSchedRequest *req;

    req =(TSchedRequest*) MALLOC(sizeof(TSchedRequest));

    req->destroyFunc = NULL;

    return req;
}

/*
 * SchedFreeRequest
 *
 * Free a scheduler request structure
 *
 * Side Effects: If there is a deallocator function for the request, it is
 *    called. This lets upper level modules (i.e. mirage-user.c) destroy
 *    any data they may have allocated.
 */

void SchedFreeRequest(TSchedRequest *req)
{
    if (req->destroyFunc != NULL) {
        req->destroyFunc(req);
    }

    FREE(req);
}

/*
 * SchedFindClient
 *
 * Search the scheduler list for a specific client ID.
 *
 */

TSchedHeader *SchedFindClient(int clientId)
{
    TSchedHeader *cur = SchedList;

    while (cur != NULL) {
        if (cur->clientId == clientId) {
            return cur;
        }
        cur = cur -> next;
    }

    return NULL;
}

/*
 * SchedEnqueue
 *
 * Given a scheduler request struct, enqueue it into the scheduler list. If
 * the circular queue for the client is full, then throw away the entry.
 *
 */

void SchedEnqueue(TSchedRequest *request)
{
    TSchedHeader *header = SchedFindClient(request->clientId);
    TSchedRequest *itemPtr;

    if (header == NULL) {
        header = (TSchedHeader*) MALLOC(sizeof(TSchedHeader));
        header->clientId = request->clientId;
        header->bankBalance = SCHED_INITIAL_BALANCE;
        header->shares = SCHED_SHARES;
        CircularInit(&header->requestQueue, SCHED_MAX_REQUESTS_PER_CLIENT);
        header->next = SchedList;
        SchedList = header;
    }

    if (CircularFull(&header->requestQueue)) {
        CircularDequeue(&header->requestQueue, FALSE, (void**) &itemPtr);
        SchedFreeRequest(itemPtr);
        SchedNumWaiting--;
        // throw it away
    }

    CircularEnqueue(&header->requestQueue, (void*) request);
    SchedNumWaiting++;
}

/*
 * SchedInitRequest
 *
 * Allocate a scheduler request function and fill in the clientID, cost, and
 * deallocator.
 *
 */

TSchedRequest *SchedInitRequest(int dosId,
    int dosScore,
    DestroyRequestFunc *destroyFunc)
{
    TSchedRequest *req;

    req = SchedAllocRequest();

    req->clientId = dosId;
    req->schedCost = dosScore;
    req->destroyFunc = destroyFunc;

    return req;
}

/*
 * SchedPopRequest
 *
 * Pop a request from the scheduler. This uses charge based proportional
 * scheduling [Umesh Maheshwari, Technical Memo MIT/LCS/TM-529, M.I.T]
 *
 */

Status SchedPopRequest(TSchedRequest **request)
{
    int runnable;

    if (SchedList == NULL) {
        return FAILURE;
    }

onceMore:
    if (RoundRobinPtr == NULL) {
        RoundRobinPtr = SchedList;
    }

    runnable = FALSE;

    while (RoundRobinPtr != NULL) {
        if (!CircularEmpty(&RoundRobinPtr->requestQueue)) {
            runnable = TRUE;
            if (RoundRobinPtr->bankBalance > 0) {
                CircularDequeue(&RoundRobinPtr->requestQueue, FALSE,
                        (void**) request);
                RoundRobinPtr->bankBalance -= (*request)->schedCost;
                SchedNumWaiting--;
                return SUCCESS;
            }
        }

        RoundRobinPtr = RoundRobinPtr->next;
    }

    if (runnable) {
        RoundRobinPtr = SchedList;
        while (RoundRobinPtr != NULL) {
            RoundRobinPtr->bankBalance += RoundRobinPtr->shares;
            if (RoundRobinPtr->bankBalance > SCHED_MAX_BALANCE) {
                RoundRobinPtr->bankBalance = SCHED_MAX_BALANCE;
            }

            RoundRobinPtr = RoundRobinPtr->next;
        }

        // something was runnable, but we didn't run it due to insufficient
        // balance, so try again
        goto onceMore;
    }

    return FAILURE;
}




