/*
 *  The Mirage NFS Router
 *
 *  @COPYRIGHT@
 *
 *  transact.c
 *
 *  NOTE: Any changes to this file should also be made to ttSilk.c !!
 *
 */

#include "nfsroute.h"

#define TT_FAST_LAST_ENTRY
#define TT_FREELIST

Hash_Table TransactHashTable;

TTransactEntry *TransactEntryFreeList;
TTransactEntry *LastTransactEntry;
TTransactKey LastTransactKey;

/*
 * TransactGetOutstanding
 *
 * Return the number of entries in the TT that are waiting to for replies.
 *
 */

int TransactGetOutstanding()
{
    return TransactHashTable.numEntries;
}

/*
 * TransactInit
 *
 * Initialize this module
 *
 */

Status TransactInit()
{
    Hash_InitTable(&TransactHashTable, 0, sizeof(TTransactKey)/4);

    LastTransactEntry = NULL;
    TransactEntryFreeList = NULL;

    return SUCCESS;
}

/*
 * TransactEntryAlloc
 *
 * Allocate a transactEntry from the freelist.
 *
 */

TTransactEntry *TransactEntryAlloc()
{
    TTransactEntry *entry;

    if (TransactEntryFreeList!=NULL) {
        entry = TransactEntryFreeList;
        TransactEntryFreeList = TransactEntryFreeList->next;
    } else {
        entry = (TTransactEntry*) MALLOC(sizeof(TTransactEntry));
    }

    return entry;
}

/*
 * TransactEntryFree
 *
 * return a transactEntry to the freelist.
 *
 */

void TransactEntryFree(TTransactEntry *entry)
{
#ifdef TT_FREELIST
    entry->next = TransactEntryFreeList;
    TransactEntryFreeList = entry;
#else
    FREE(entry);
#endif
}

/*
 * TransactCreateKey
 *
 * Create a key that can be used by TransactCreate() and TransactLookup()
 *
 */

Status TransactCreateKey(int xid, int clientIp, TTransactKey *key)
{
    key->xid = xid;
    key->clientIp = clientIp;

    return SUCCESS;
}

/*
 * TransactAddToHash
 *
 * Add a transaction entry to the hash table.
 *
 */

Status TransactAddToHash(TTransactKey *key, TTransactEntry *entry)
{
    Hash_Entry *h;
    Boolean isNew;

    h = Hash_CreateEntry(&TransactHashTable, (void*) key, &isNew);
    Hash_SetValue(h, entry);
    entry->h = h;

    return SUCCESS;
}

/*
 * TransactFillInvalidDosFields
 *
 * Fill in fields pertaining to the invalid DoS checking. OBSOLETE
 *
 */

#ifdef TT_HAS_INVALID_DOS_FIELDS
Status TransactFillInvalidDosFields(TTransactEntry *entry,
    char *filename,
    int pvin,
    int vin2,
    char *filename2)
{
    entry->pvin = pvin;

    if ((filename) && (strlen(filename) < MAX_LOOKUP_FILENAME))  {
        strcpy(entry->filename, filename);
    }

    entry->vin2 = vin2;

    if ((filename2) && (strlen(filename2) < MAX_LOOKUP_FILENAME))  {
        strcpy(entry->filename2, filename2);
    }

    return SUCCESS;
}
#endif

/*
 * TransactPurge
 *
 * Purge old entries from the transaction table.
 *
 */

void TransactPurge(int maxAge)
{
    int time = TimeGetS() - maxAge;

    Hash_Entry *h;
    Hash_Search hashSearchPtr;

    h = Hash_EnumFirst(&TransactHashTable,&hashSearchPtr);
    while (h!=NULL) {
        TTransactEntry *transactEntry = (TTransactEntry*) Hash_GetValue(h);

        if (transactEntry->timeStamp < time) {
            TransactEntryFree(transactEntry);
            Hash_DeleteEntry(&TransactHashTable, h);
        }
        h = Hash_EnumNext(&hashSearchPtr);
    }
}

/*
 * TransactCreate
 *
 * Create a new transaction entry.
 *
 * A special one-entry buffer is kept for the
 * current transaction entry in the hope that it can be serviced without
 * needing to be put in the hash table. If a new entry arrives, and the entry
 * waiting in the buffer still has not been served, then the entry in the buffer
 * will be put into the hash table.
 *
 */

Status TransactCreate(TTransactKey *key,
    int proc,
    int vin,
    int mountChecksum,
    int uid,
    int gid,
    int dosId,

    TTransactEntry **entryOut)
{
    TTransactEntry *entry;

    entry = TransactEntryAlloc();

    entry->timeStamp = TimeGetS();
    entry->proc = proc;
    entry->vin = vin;
    entry->mountChecksum = mountChecksum;
    entry->uid = uid;
    entry->gid = gid;
    entry->dosId = dosId;
    entry->refCount = 1;
    entry->h = NULL;

#ifdef TT_HAS_INVALID_DOS_FIELDS
    // invalid dos fields
    entry->pvin = -1;
    *entry->filename = 0;
    entry->vin2 = -1;
    *entry->filename2 = 0;
#endif

#ifdef TT_FAST_LAST_ENTRY
    if (LastTransactEntry != NULL) {
        TransactAddToHash(&LastTransactKey, LastTransactEntry);
    }
    LastTransactEntry = entry;
    LastTransactKey = *key;
#else
    TransactAddToHash(key, entry);
#endif

    if (entryOut) {
        *entryOut = entry;
    }

    return SUCCESS;
}

/*
 * TransactFree
 *
 * Remove an entry from the TT and free it.
 *
 */

Status TransactFree(TTransactEntry *entry)
{
    entry->refCount--;
    if (entry->refCount<=0) {
        Hash_Entry *h = entry->h;

        TransactEntryFree(entry);

        if (h!= NULL) {
            Hash_DeleteEntry(&TransactHashTable, h);
        }

        if (LastTransactEntry == entry) {
            LastTransactEntry = NULL;
        }
    }

    return SUCCESS;
}

/*
 * TranactLookup
 *
 * Search the TT for the key and return the entry if it exists.
 *
 */

Status TransactLookup(TTransactKey *key, TTransactEntry **entry)
{
    Hash_Entry *h;

    if ((LastTransactEntry != NULL) &&
            (memcmp(key, &LastTransactKey, sizeof(TTransactKey))==0)) {
        *entry = LastTransactEntry;
        return SUCCESS;
    }

    h = Hash_FindEntry(&TransactHashTable, (void*) key);
    if (h == NULL) {
        return FAILURE;
    }

    (*entry) = (TTransactEntry*) Hash_GetValue(h);
    if (*entry == NULL) {
        return FAILURE;
    }

    return SUCCESS;
}


