/*
 *  The Mirage NFS Router
 *
 *  @COPYRIGHT@
 *
 *  umt.c: user mapping table
 *
 *  NOTE: Any changes to this file should also be made to umtSilk.c !!
 *
 */

#include "nfsroute.h"

#define UMT_FAST_LAST_ENTRY
#define UMT_FREELIST

int UmtXidCounter = 0;
Hash_Table UmtHashTable;

TUmtEntry *UmtFreeList;
TUmtEntry *LastUmtEntry;
int LastUmtXid;

/*
 * UmtInit
 *
 * Initialize this module
 *
 */

Status UmtInit()
{
    Hash_InitTable(&UmtHashTable, 0, HASH_ONE_WORD_KEYS);
    LastUmtEntry = NULL;
    UmtFreeList = NULL;

    return SUCCESS;
}

/*
 * UmtEntryAlloc
 *
 * Allocate an entry from the freelist.
 *
 */

TUmtEntry *UmtEntryAlloc()
{
    TUmtEntry *entry;

    if (UmtFreeList!=NULL) {
        entry = UmtFreeList;
        UmtFreeList = UmtFreeList->next;
    } else {
        entry = (TUmtEntry*) MALLOC(sizeof(TUmtEntry));
    }

    return entry;
}

/*
 * UmtEntryFree
 *
 * Return an entry to the freelist.
 *
 */

void UmtEntryFree(TUmtEntry *entry)
{
#ifdef UMT_FREELIST
    entry->next = UmtFreeList;
    UmtFreeList = entry;
#else
    FREE(entry);
#endif
}

/*
 * UmtAddToHash
 *
 * Add a UMT entry to the UMT hash table
 *
 */

Status UmtAddToHash(int xid, TUmtEntry *umtEntry)
{
    Hash_Entry *h;
    Boolean isNew;

    h = Hash_CreateEntry(&UmtHashTable, (void*) xid, &isNew);
    Hash_SetValue(h, umtEntry);

    umtEntry->h = h;

    return SUCCESS;
}

/*
 * UmtCreate
 *
 * Create a new UMT entry
 *
 * Side Effects: newXid is initialized with the new Xid
 *
 */

Status UmtCreate(int oldXid, int clientIp, int clientPort, int *newXid,
    TUmtEntry **umtEntryOut)
{
    TUmtEntry *umtEntry;

    *newXid = UmtXidCounter++;

    umtEntry = UmtEntryAlloc();
    umtEntry->oldXid = oldXid;
    umtEntry->clientIp = clientIp;
    umtEntry->clientPort = clientPort;
    umtEntry->refCount = 1;
    umtEntry->h = NULL;

#ifdef UMT_FAST_LAST_ENTRY
    if (LastUmtEntry != NULL) {
        UmtAddToHash(LastUmtXid, LastUmtEntry);
    }
    LastUmtXid = *newXid;
    LastUmtEntry = umtEntry;
#else
    UmtAddToHash(*newXid, umtEntry);
#endif

    if (umtEntryOut) {
        *umtEntryOut = umtEntry;
    }

    return SUCCESS;
}

/*
 * UmtLookup
 *
 * Search the UMT for an Xid, and return the corresponding Umt entry
 *
 */

Status UmtLookup(int newXid, TUmtEntry **umtEntryOut)
{
    Hash_Entry *h;
    TUmtEntry *umtEntry;

    if ((LastUmtEntry!=NULL) && (LastUmtXid == newXid)) {
        *umtEntryOut = LastUmtEntry;
        return SUCCESS;
    }

    h = Hash_FindEntry(&UmtHashTable, (void*) newXid);
    if (h == NULL) {
        return FAILURE;
    }

    umtEntry = (TUmtEntry*) Hash_GetValue(h);
    if (umtEntry == NULL) {
        return FAILURE;
    }

    if (umtEntryOut) {
        *umtEntryOut = umtEntry;
    }

    return SUCCESS;
}

Status UmtFree(TUmtEntry *umtEntry)
{
    umtEntry->refCount--;
    if (umtEntry->refCount <= 0) {
        Hash_Entry *h = umtEntry->h;
        UmtEntryFree(umtEntry);

        if (h != NULL) {
            Hash_DeleteEntry(&UmtHashTable, h);
        }

        if (LastUmtEntry == umtEntry) {
            LastUmtEntry = NULL;
        }
    }

    return SUCCESS;
}



